"""Stereo Calibration module"""
import glob
import os
import math
import cv2
import numpy as np


class StereoCalibration():
    def __init__(self):
        self.flags = 0
        self.flags |= cv2.CALIB_FIX_INTRINSIC
        # self.flags |= cv2.CALIB_FIX_PRINCIPAL_POINT
        self.flags |= cv2.CALIB_USE_INTRINSIC_GUESS
        self.flags |= cv2.CALIB_FIX_FOCAL_LENGTH
        # self.flags |= cv2.CALIB_FIX_ASPECT_RATIO
        self.flags |= cv2.CALIB_ZERO_TANGENT_DIST
        # self.flags |= cv2.CALIB_RATIONAL_MODEL
        # self.flags |= cv2.CALIB_SAME_FOCAL_LENGTH
        # self.flags |= cv2.CALIB_FIX_K3
        # self.flags |= cv2.CALIB_FIX_K4
        # self.flags |= cv2.CALIB_FIX_K5

        self.criteria = (cv2.TERM_CRITERIA_EPS +
                         cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
        self.stereocalib_criteria = (cv2.TERM_CRITERIA_MAX_ITER +
                                     cv2.TERM_CRITERIA_EPS, 100, 1e-5)
        self.stereo_cal = None
        self.cv_window_name = "Calibration"

    def read_images(self, folder, wildcard):
        # TODO check folder exists
        images = glob.glob(os.path.join(folder, wildcard))
        images.sort()
        images_cv = []
        for fname in images:
            img = cv2.imread(fname, cv2.IMREAD_COLOR)
            gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            images_cv.append(gray)
        return images_cv

    def write_yaml_calibration(self, output_folder):
        # write yamls
        left_cal_yml_file = output_folder + "/left.yaml"
        right_cal_yml_file = output_folder + "/right.yaml"
        fs_l = cv2.FileStorage(left_cal_yml_file, cv2.FILE_STORAGE_WRITE)
        fs_l.write("image_width", self.stereo_cal["image size"][0])
        fs_l.write("image_height", self.stereo_cal["image size"][1])
        fs_l.write("camera_name", "leftCamera")
        fs_l.write("camera_matrix", self.stereo_cal["left"]["m"])
        fs_l.write("distortion_model", "plumb_bob")
        fs_l.write("distortion_coefficients", self.stereo_cal["left"]["d"])
        fs_l.write("rectification_matrix", self.stereo_cal["left"]["r"])
        fs_l.write("projection_matrix", self.stereo_cal["left"]["p"])
        # rms error for stereo reprojection as
        # yaml doesn't have seperate stereo calibration file
        fs_l.write("rms_error", self.stereo_cal["error"])
        fs_l.release()
        fs_r = cv2.FileStorage(right_cal_yml_file, cv2.FILE_STORAGE_WRITE)
        fs_r.write("image_width", self.stereo_cal["image size"][0])
        fs_r.write("image_height", self.stereo_cal["image size"][1])
        fs_r.write("camera_name", "rightCamera")
        fs_r.write("camera_matrix", self.stereo_cal["right"]["m"])
        fs_r.write("distortion_model", "plumb_bob")
        fs_r.write("distortion_coefficients", self.stereo_cal["right"]["d"])
        fs_r.write("rectification_matrix", self.stereo_cal["right"]["r"])
        fs_r.write("projection_matrix", self.stereo_cal["right"]["p"])
        # rms error for stereo reprojection as
        # yaml doesn't have seperate stereo calibration file
        fs_r.write("rms_error", self.stereo_cal["error"])
        fs_r.release()

    def write_xml_calibration(self, output_folder):
        # write calibration xmls
        left_cal_xml_file = output_folder + "/left_calibration.xml"
        right_cal_xml_file = output_folder + "/right_calibration.xml"
        stereo_cal_xml_file = output_folder + "/stereo_calibration.xml"
        fs_l = cv2.FileStorage(left_cal_xml_file, cv2.FILE_STORAGE_WRITE)
        fs_l.write("cameraMatrix", self.stereo_cal["left"]["m"])
        fs_l.write("distCoeffs", self.stereo_cal["left"]["d"])
        fs_l.write("rms_error", self.stereo_cal["left"]["error"])
        fs_l.release()
        fs_r = cv2.FileStorage(right_cal_xml_file, cv2.FILE_STORAGE_WRITE)
        fs_r.write("cameraMatrix", self.stereo_cal["right"]["m"])
        fs_r.write("distCoeffs", self.stereo_cal["right"]["d"])
        fs_r.write("rms_error", self.stereo_cal["right"]["error"])
        fs_r.release()
        fs_s = cv2.FileStorage(stereo_cal_xml_file, cv2.FILE_STORAGE_WRITE)
        fs_s.write("R", self.stereo_cal["r"])
        fs_s.write("T", self.stereo_cal["t"])
        fs_s.write("Q", self.stereo_cal["q"])
        fs_s.write("E", self.stereo_cal["e"])
        fs_s.write("F", self.stereo_cal["f"])
        fs_s.write("rms_error", self.stereo_cal["error"])
        fs_s.release()
        # write rectification xmls
        left_rect_xml_file = output_folder + "/left_rectification.xml"
        right_rect_xml_file = output_folder + "/right_rectification.xml"
        fs_l = cv2.FileStorage(left_rect_xml_file, cv2.FILE_STORAGE_WRITE)
        fs_l.write("x", self.stereo_cal["left"]["map1"])
        fs_l.write("y", self.stereo_cal["left"]["map2"])
        fs_l.release()
        fs_r = cv2.FileStorage(right_rect_xml_file, cv2.FILE_STORAGE_WRITE)
        fs_r.write("x", self.stereo_cal["right"]["map1"])
        fs_r.write("y", self.stereo_cal["right"]["map2"])
        fs_r.release()

    def write_calibration(self, output_folder):
        print("Writing calibration to output folder...")
        self.write_yaml_calibration(output_folder)
        self.write_xml_calibration(output_folder)
        print("Calibration written to output folder.")

    def get_cal_from_ideal(self, resolution, pixel_pitch, focal_length,
                           baseline, output_folder=None):
        # calculate calibration values
        sensorSize = [pixel_pitch*resolution[0], pixel_pitch*resolution[1]]
        v = [sensorSize[0]/(2*focal_length), sensorSize[0]/(2*focal_length)]
        FOV = [2*math.atan(v[0]), 2*math.atan(v[1])]
        focalLength_px = [(resolution[0]/2)/math.tan(FOV[0]/2),
                          (resolution[1]/2)/math.tan(FOV[1]/2)]
        c = [resolution[0]/2, resolution[1]/2]

        # TODO find out why focal length when calibrating in
        # the real world only uses x focal length
        focalLength_px[1] = focalLength_px[0]

        # assign values to calibration matrices
        m_l = np.asarray([(focalLength_px[0], 0., c[0]),
                          (0., focalLength_px[1], c[1]),
                          (0., 0., 1.)])
        m_r = np.asarray([(focalLength_px[0], 0., c[0]),
                         (0., focalLength_px[1], c[1]),
                         (0., 0., 1.)])
        d_l = np.asarray([0, 0, 0, 0, 0])
        d_r = np.asarray([0, 0, 0, 0, 0])
        r_l = np.asarray([(1, 0, 0),
                          (0, 1, 0),
                          (0, 0, 1)])
        r_r = np.asarray([(1, 0, 0),
                          (0, 1, 0),
                          (0, 0, 1)])
        p_l = np.asarray([(focalLength_px[0], 0., c[0], 0.),
                          (0., focalLength_px[1], c[1], 0.),
                          (0., 0., 1., 0.)])
        p_r = np.asarray([
            (focalLength_px[0], 0., c[0], -focalLength_px[0]*baseline),
            (0., focalLength_px[1], c[1], 0.),
            (0., 0., 1., 0.)])

        q = self.calc_q(m_l, p_r, p_l)

        # missing data from calibration files
        # (not possible to know if just using yaml)
        camera_l_cal = {
            'm': m_l, 'd': d_l, 'r': r_l, 'p': p_l,
            'map1': None, 'map2': None, 'error': None}
        camera_r_cal = {
            'm': m_r, 'd': d_r, 'r': r_r, 'p': p_r,
            'map1': None, 'map2': None, 'error': None}
        stereo_cal = {
            'image size': resolution,
            'left': camera_l_cal, 'right': camera_r_cal,
            'q': q, 'r': None, 't': None, 'e': None, 'f': None, 'error': None}

        self.stereo_cal = stereo_cal
        if output_folder is not None:
            self.write_yaml_calibration(output_folder)

        return True, stereo_cal

    def get_cal_from_yaml(self, left_cal_file, right_cal_file):
        res = False
        fs_l = cv2.FileStorage(left_cal_file, cv2.FILE_STORAGE_READ)
        fs_r = cv2.FileStorage(right_cal_file, cv2.FILE_STORAGE_READ)
        if fs_l.isOpened() and fs_r.isOpened():
            m_l = fs_l.getNode("camera_matrix").mat()
            m_r = fs_r.getNode("camera_matrix").mat()
            d_l = fs_l.getNode("distortion_coefficients").mat()
            d_r = fs_r.getNode("distortion_coefficients").mat()
            r_l = fs_l.getNode("rectification_matrix").mat()
            r_r = fs_r.getNode("rectification_matrix").mat()
            p_l = fs_l.getNode("projection_matrix").mat()
            p_r = fs_r.getNode("projection_matrix").mat()
            res = True
        else:
            print("Failed to open calibration files")
            return False, None
        fs_l.release()
        fs_r.release()

        q = self.calc_q(m_l, p_r, p_l)

        # missing data from calibration files
        # (not possible to know if just using yaml)
        camera_l_cal = {
            'm': m_l, 'd': d_l, 'r': r_l, 'p': p_l,
            'map1': None, 'map2': None, 'error': None}
        camera_r_cal = {
            'm': m_r, 'd': d_r, 'r': r_r, 'p': p_r,
            'map1': None, 'map2': None, 'error': None}
        stereo_cal = {
            'image size': None,
            'left': camera_l_cal, 'right': camera_r_cal,
            'q': q, 'r': None, 't': None, 'e': None, 'f': None, 'error': None}

        self.stereo_cal = stereo_cal
        return res, stereo_cal

    def get_cal_from_xml(self, left_cal_file, right_cal_file, stereo_cal_file,
                         left_rect_file, right_rect_file):
        res = False
        fs_l = cv2.FileStorage(left_cal_file, cv2.FILE_STORAGE_READ)
        fs_r = cv2.FileStorage(right_cal_file, cv2.FILE_STORAGE_READ)
        fs_s = cv2.FileStorage(stereo_cal_file, cv2.FILE_STORAGE_READ)
        fs_r_l = cv2.FileStorage(left_rect_file, cv2.FILE_STORAGE_READ)
        fs_r_r = cv2.FileStorage(right_rect_file, cv2.FILE_STORAGE_READ)
        if fs_l.isOpened() and fs_r.isOpened() and fs_s.isOpened() \
           and fs_r_l.isOpened() and fs_r_r.isOpened():
            m_l = fs_l.getNode("cameraMatrix").mat()
            m_r = fs_r.getNode("cameraMatrix").mat()
            d_l = fs_l.getNode("distCoeffs").mat()
            d_r = fs_r.getNode("distCoeffs").mat()
            error_l = fs_l.getNode("rms_error").real()
            error_r = fs_r.getNode("rms_error").real()
            r_s = fs_s.getNode("R").mat()
            t_s = fs_s.getNode("T").mat()
            q_s = fs_s.getNode("Q").mat()
            e_s = fs_s.getNode("E").mat()
            f_s = fs_s.getNode("F").mat()
            error_s = fs_s.getNode("rms_error").real()
            mapL1 = fs_r_l.getNode("x").mat()
            mapL2 = fs_r_l.getNode("y").mat()
            mapR1 = fs_r_r.getNode("x").mat()
            mapR2 = fs_r_r.getNode("y").mat()
            res = True
        else:
            print("Failed to open calibration files")
            return False, None
        fs_l.release()
        fs_r.release()
        fs_s.release()
        fs_r_l.release()
        fs_r_r.release()

        # missing data from calibration files
        # (not possible to know if just using xml)
        camera_l_cal = {
            'm': m_l, 'd': d_l, 'r': None, 'p': None,
            'map1': mapL1, 'map2': mapL2, 'error': error_l}
        camera_r_cal = {
            'm': m_r, 'd': d_r, 'r': None, 'p': None,
            'map1': mapR1, 'map2': mapR2, 'error': error_r}
        stereo_cal = {
            'image size': None,
            'left': camera_l_cal, 'right': camera_r_cal,
            'q': q_s, 'r': r_s, 't': t_s, 'e': e_s, 'f': f_s, 'error': error_s}

        self.stereo_cal = stereo_cal
        return res, stereo_cal

    def calc_q(self, m_l, p_r, p_l):
        q = np.zeros(shape=(4, 4))
        self.cx = p_l[0, 2]
        self.cxr = p_r[0, 2]
        self.cy = p_l[1, 2]
        self.fx = m_l[0, 0]
        self.fy = m_l[1, 1]

        p14 = p_r[0, 3]
        self.T = -p14 / self.fx

        q33 = -(self.cx - self.cxr)/self.T

        # calucalte Q values
        q[0, 0] = 1.0
        q[0, 3] = -self.cx
        q[1, 1] = 1.0
        q[1, 3] = -self.cy
        q[2, 3] = self.fx
        q[3, 2] = 1.0 / self.T
        q[3, 3] = q33

        return q

    def calibrate_camera(self, images, grid_rows, grid_cols, objp):
        objpoints = []  # 3d point in real world space
        imgpoints = []  # 2d points in image plane.

        for img in images:
            ret, corners = cv2.findChessboardCorners(
                img, (grid_cols, grid_rows), None)
            if ret:
                objpoints.append(objp)
                ret = cv2.cornerSubPix(
                    img, corners, (11, 11), (-1, -1), self.criteria)
                imgpoints.append(corners)

        img_shape = images[0].shape[::-1]

        ret, m, d, r, p = cv2.calibrateCamera(
            objpoints, imgpoints, img_shape, None, None)

        camera_l_cal = {
            'imgpoints': imgpoints, 'm': m, 'd': d, 'r': r, 'p': p}

        return objpoints, camera_l_cal

    def calibrate_cameras(self, images_l, images_r,
                          grid_rows, grid_cols, objp):
        objpoints = []  # 3d point in real world space
        imgpoints_l = []  # 2d points in image plane.
        imgpoints_r = []  # 2d points in image plane.

        cv2.namedWindow(self.cv_window_name)

        for img_l, img_r in zip(images_l, images_r):
            ret_l, corners_l = cv2.findChessboardCorners(
                img_l, (grid_cols, grid_rows), None)
            ret_r, corners_r = cv2.findChessboardCorners(
                img_r, (grid_cols, grid_rows), None)
            if (ret_l and ret_r):
                objpoints.append(objp)
                corners_l = cv2.cornerSubPix(
                    img_l, corners_l, (11, 11), (-1, -1), self.criteria)
                corners_r = cv2.cornerSubPix(
                    img_r, corners_r, (11, 11), (-1, -1), self.criteria)
                imgpoints_l.append(corners_l)
                imgpoints_r.append(corners_r)

                ret_l = cv2.drawChessboardCorners(
                    img_l, (grid_cols, grid_rows),
                    corners_l, ret_l)
                ret_r = cv2.drawChessboardCorners(
                    img_r, (grid_cols, grid_rows),
                    corners_r, ret_r)

                image_left_resized = cv2.resize(img_l, (640, 480))
                image_right_resized = cv2.resize(img_r, (640, 480))

                left_right_dual = np.concatenate(
                    (image_left_resized, image_right_resized), axis=1)

                cv2.imshow(self.cv_window_name, left_right_dual)
                cv2.waitKey(1)

        cv2.destroyWindow(self.cv_window_name)

        img_shape = images_l[0].shape[::-1]

        ret_l, m_l, d_l, r_l, t_l = cv2.calibrateCamera(
            objpoints, imgpoints_l, img_shape, None, None)
        ret_r, m_r, d_r, r_r, t_r = cv2.calibrateCamera(
            objpoints, imgpoints_r, img_shape, None, None)

        camera_l_cal = {
            'imgpoints': imgpoints_l,
            'm': m_l, 'd': d_l, 'r': r_l, 't': t_l,
            'error': ret_l}
        camera_r_cal = {
            'imgpoints': imgpoints_r,
            'm': m_r, 'd': d_r, 'r': r_r, 't': t_r,
            'error': ret_r}

        return objpoints, camera_l_cal, camera_r_cal

    def calibrate(self, left_images_folder, right_images_folder, output_folder,
                  left_wildcard="*_l", right_wildcard="*_r",
                  grid_size=39.0, grid_rows=6, grid_cols=9):
        print("Calibrating stereo images...")

        self.grid_size = grid_size

        objp = np.zeros((grid_rows*grid_cols, 3), np.float32)
        objp[:, :2] = np.mgrid[0:grid_cols, 0:grid_rows].T.reshape(-1, 2)

        images_left = self.read_images(left_images_folder, left_wildcard)
        images_right = self.read_images(right_images_folder, right_wildcard)
        objpoints, camera_l_cal, camera_r_cal = self.calibrate_cameras(
            images_left, images_right, grid_rows, grid_cols, objp)

        img_shape = images_left[0].shape[::-1]

        ret, _, _, _, _, R, T, E, F = cv2.stereoCalibrate(
            objpoints, camera_l_cal["imgpoints"], camera_r_cal["imgpoints"],
            camera_l_cal["m"], camera_l_cal["d"],
            camera_r_cal["m"], camera_r_cal["d"],
            img_shape, criteria=self.stereocalib_criteria, flags=self.flags)

        R1, R2, P1, P2, Q, _, _ = cv2.stereoRectify(
            camera_l_cal["m"], camera_l_cal["d"],
            camera_r_cal["m"], camera_r_cal["d"],
            img_shape, R, T)

        stcamera_l_cal = {
            'm': camera_l_cal["m"], 'd': camera_l_cal["d"],
            'r': R1, 'p': P1, 'map1': None, 'map2': None,
            'error': camera_l_cal["error"]}
        stcamera_r_cal = {
            'm': camera_r_cal["m"], 'd': camera_r_cal["d"],
            'r': R2, 'p': P2, 'map1': None, 'map2': None,
            'error': camera_r_cal["error"]}

        mapL1, mapL2 = cv2.initUndistortRectifyMap(
                stcamera_l_cal["m"], stcamera_l_cal["d"],
                stcamera_l_cal["r"], stcamera_l_cal["p"],
                img_shape, cv2.CV_32FC1)

        mapR1, mapR2 = cv2.initUndistortRectifyMap(
                stcamera_r_cal["m"], stcamera_r_cal["d"],
                stcamera_r_cal["r"], stcamera_r_cal["p"],
                img_shape, cv2.CV_32FC1)

        stcamera_l_cal["map1"] = mapL1
        stcamera_l_cal["map2"] = mapL2
        stcamera_r_cal["map1"] = mapR1
        stcamera_r_cal["map2"] = mapR2

        stereo_cal = {
            'image size': img_shape,
            'left': stcamera_l_cal, 'right': stcamera_r_cal,
            'q': Q, 'r': R, 't': T, 'e': E, 'f': F, 'error': ret}

        self.stereo_cal = stereo_cal

        self.write_calibration(output_folder)

        print("Calibration complete")
        return stereo_cal

    def rectify_pair(self, image_left, image_right):
        image_l_rect = self.rectify(image_left, self.stereo_cal["left"])
        image_r_rect = self.rectify(image_right, self.stereo_cal["right"])
        return image_l_rect, image_r_rect

    def rectify(self, image, camera_cal):
        if (camera_cal["map1"] is None or camera_cal["map2"] is None):
            # create rectification maps if not read from calibration file
            map1, map2 = cv2.initUndistortRectifyMap(
                camera_cal["m"], camera_cal["d"],
                camera_cal["r"], camera_cal["p"],
                image.shape[::-1], cv2.CV_32FC1)
            camera_cal["map1"] = map1
            camera_cal["map2"] = map2
        else:
            map1 = camera_cal["map1"]
            map2 = camera_cal["map2"]
        rect_image = cv2.remap(
            image, map1, map2, cv2.INTER_CUBIC, borderMode=cv2.BORDER_CONSTANT)
        return rect_image



import cv2
import numpy as np


class CVImageCapture():
    def __init__(self, filename):
        """
        Initialisation function for CVCapture class.
        :param filename: filename of the image to read
        :type filename: string
        """
        self.image = None
        self.filename = filename

    def connect(self):
        """
        Connect to camera using opencv VideoCaputre
        :returns: success of connection
        :rtype: bool
        """
        self.image = cv2.imread(self.filename, cv2.IMREAD_GRAYSCALE)
        return True

    def grab(self):
        """
        Grab image from camera
        :returns: success of capture, image
        :rtype: bool, numpy.array
        """
        res = not np.shape(self.image) == ()
        return res, self.image

    def close(self):
        """
        Close connection
        """


if __name__ == '__main__':
    FILENAME = "test.png"
    cam = CVImageCapture(FILENAME)
    cam.connect()
    while(True):
        res, image = cam.grab()
        if (res):
            image_resized = cv2.resize(image, (640, 480))
            cv2.imshow('Image', image_resized)
            k = cv2.waitKey(1)
            if k == ord('q'):
                break
    cam.close()



from stereo3d.stereocapture.coppeliasim import VREPConnection, VREPStereoVisionSensor
import numpy as np
import cv2


class StereoCaptureVREP():
    def __init__(self, left_camera_name, right_camera_name, api_port):
        """
        Initialisation function for StereoCaptureVREP class.
        :param camL: left camera
        :param camR: right camera
        :type camL: PylonCapture
        :type camR: PylonCapture
        """
        self.vrep_connection = None
        self.left_camera_name = left_camera_name
        self.right_camera_name = right_camera_name
        self.api_port = api_port
        self.camera = None

    def connect(self):
        """
        Connect to cameras.
        :returns: success of connection
        :rtype: bool
        """
        self.vrep_connection = VREPConnection(self.api_port)
        res, clientID = self.vrep_connection.connect()
        if (res):
            self.camera = VREPStereoVisionSensor(
                self.left_camera_name, self.right_camera_name, clientID)
            return self.camera.connect()
        else:
            return False

    def grab(self):
        """
        Grab images from each camera in stereo pair
        :returns: success of capture, image left, image right
        :rtype: bool, numpy.array, numpy.array
        """
        if (self.camera.isConnected()):
            res, image_l, image_r = self.camera.capture()
            if (res):
                image_l_mono = cv2.cvtColor(image_l, cv2.COLOR_BGR2GRAY)
                image_r_mono = cv2.cvtColor(image_r, cv2.COLOR_BGR2GRAY)
                return res, image_l_mono, image_r_mono
            else:
                False, None, None
        else:
            return False, None, None

    def close(self):
        """
        Close connection to cameras
        """
        self.camera.close()


if __name__ == "__main__":
    API_PORT = 20000
    LEFT_CAMERA_NAME = "StereoCameraLeft"
    RIGHT_CAMERA_NAME = "StereoCameraRight"
    stcam = StereoCaptureVREP(LEFT_CAMERA_NAME, RIGHT_CAMERA_NAME, API_PORT)
    stcam.connect()
    while True:
        res, imageL, imageR = stcam.grab()
        if res:
            stereo_image = np.concatenate((imageL, imageR), axis=1)
            stereo_image_resized = cv2.resize(stereo_image, (1280, 480))
            cv2.imshow('Stereo Image', stereo_image_resized)
            k = cv2.waitKey(1)
            if k == ord('q'):
                break
    stcam.close()



# Make sure to have the server side running in V-REP:
# in a child script of a V-REP scene, add following command
# to be executed just once, at simulation start:
#
# simRemoteApi.start(19999)
#
# then start simulation, and run this program.
#
# IMPORTANT: for each successful call to simxStart, there
# should be a corresponding call to simxFinish at the end!

try:
    from stereo3d.stereocapture.coppeliasim import sim as vrep
except:
    print('--------------------------------------------------------------')
    print('"sim.py" could not be imported. This means very probably that')
    print('either "sim.py" or the remoteApi library could not be found.')
    print('Make sure both are in the same folder as this file,')
    print('or appropriately adjust the file "sim.py"')
    print('--------------------------------------------------------------')
    print('')

import cv2
import numpy as np
import threading

# script_path = os.path.dirname(os.path.realpath(__file__))


class VREPConnection:
    def __init__(self, port=19999, ip='127.0.0.1'):
        self.port = port
        self.ip = ip

    def vrep_connect(self):
        clientID = vrep.simxStart(
            self.ip, self.port, True, True, 5000, 1)  # Connect to V-REP
        if clientID != -1:
            print('Connected to remote API server')
            return True, clientID
        else:
            print('Failed to connect to API server')
            return False, clientID

    def connect(self, retry=False):
        if (retry):
            connected = False
            while (not connected):
                connected, self.clientID = self.vrep_connect()
            print("Connected with clientID: {}".format(self.clientID))
            return connected, self.clientID
        else:
            connected, self.clientID = self.vrep_connect()
            print("Connected with clientID: {}".format(self.clientID))
            return connected, self.clientID

    def prep_object_pose(self, object_handle, relative_to_handle):
        res1, position = vrep.simxGetObjectPosition(
            self.clientID, object_handle, relative_to_handle, vrep.simx_opmode_streaming)
        res2, orienation = vrep.simxGetObjectOrientation(
            self.clientID, object_handle, relative_to_handle, vrep.simx_opmode_streaming)
        if (res1 == 1 and res2 == 1):
            return True
        else:
            print("Failed to setup object pose stream")
            print("{},{}".format(res1, res2))
            return False

    def send_int_signal(self, signal_name, value):
        vrep.simxSetIntegerSignal(
            self.clientID, signal_name, value, vrep.simx_opmode_oneshot_wait)

    def get_object_pose(self, object_handle, relative_to_handle):
        res = False
        res1, position = vrep.simxGetObjectPosition(
            self.clientID, object_handle, relative_to_handle, vrep.simx_opmode_buffer)
        res2, orienation = vrep.simxGetObjectOrientation(
            self.clientID, object_handle, relative_to_handle, vrep.simx_opmode_buffer)
        if (res1 == vrep.simx_return_ok and res2 == vrep.simx_return_ok):
            res = True
        return res, position, orienation


class VREPStereoVisionSensor:
    def __init__(self, left_name, right_name, clientID):
        self.left_name = left_name
        self.right_name = right_name
        self.clientID = clientID
        self.cv_image_l = None
        self.cv_image_r = None
        self.running = False
        self.position = None
        self.orientation = None

    def VREP_image_to_cv_rgb(self, h, w, img):
        cv_image = np.array(img, dtype=np.uint8)
        cv_image.resize([h, w, 3])
        cv_image = cv2.flip(cv_image, 0)
        cv_image = cv2.cvtColor(cv_image, cv2.COLOR_BGR2RGB)
        return cv_image

    def VREP_depth_to_np(self, h, w, img):
        cv_image = np.array(img, dtype=np.float32)
        cv_image.resize([h, w])
        cv_image = cv2.flip(cv_image, 0)
        return cv_image

    def disconnect(self):
        # Before closing the connection to V-REP, make sure that the last command sent out had time to arrive.
        vrep.simxGetPingTime(self.clientID)

        # Now close the connection to V-REP:
        vrep.simxFinish(self.clientID)

    def connect(self):
        if self.clientID != -1:
            #print ('Connected to remote API server')
            res_l, self.left_handle = vrep.simxGetObjectHandle(
                self.clientID, self.left_name, vrep.simx_opmode_oneshot_wait)
            res_r, self.right_handle = vrep.simxGetObjectHandle(
                self.clientID, self.right_name, vrep.simx_opmode_oneshot_wait)
            if (res_l == vrep.simx_return_ok and res_r == vrep.simx_return_ok):
                vrep.simxPauseCommunication(self.clientID, 1)
                res_l, self.resolution, self.image_l = vrep.simxGetVisionSensorImage(
                    self.clientID, self.left_handle, 0, vrep.simx_opmode_streaming)
                res_r, self.resolution, self.image_r = vrep.simxGetVisionSensorImage(
                    self.clientID, self.right_handle, 0, vrep.simx_opmode_streaming)
                res_p, self.position = vrep.simxGetObjectPosition(
                    self.clientID, self.left_handle, -1, vrep.simx_opmode_streaming)
                res_o, self.orientation = vrep.simxGetObjectOrientation(
                    self.clientID, self.left_handle, -1, vrep.simx_opmode_streaming)
                vrep.simxPauseCommunication(self.clientID, 0)

                if (res_l == 1 and res_r == 1 and res_p == 1 and res_o == 1):
                    return True
                else:
                    print("Failed to read vision sensors")
                    print("{},{},{},{}".format(res_l, res_r, res_p, res_o))
                    return False
            else:
                print("Failed to connect to vision sensor objects: {},{}".format(
                    res_l, res_r))
                return False
        else:
            print("Failed to connect VREP")
            return False

    def isConnected(self):
        if (vrep.simxGetConnectionId(self.clientID) != -1):
            return True
        else:
            return False

    def isInitalised(self):
        return self.initalised

    def capture(self):
        if (vrep.simxGetConnectionId(self.clientID) != -1):
            #print("Getting image from VREP...")
            vrep.simxPauseCommunication(self.clientID, 1)
            res_l, self.resolution, self.image_l = vrep.simxGetVisionSensorImage(
                self.clientID, self.left_handle, 0, vrep.simx_opmode_buffer)
            res_r, self.resolution, self.image_r = vrep.simxGetVisionSensorImage(
                self.clientID, self.right_handle, 0, vrep.simx_opmode_buffer)
            res_p, self.position = vrep.simxGetObjectPosition(
                self.clientID, self.left_handle, -1, vrep.simx_opmode_buffer)
            res_o, self.orientation = vrep.simxGetObjectOrientation(
                self.clientID, self.left_handle, -1, vrep.simx_opmode_buffer)
            vrep.simxPauseCommunication(self.clientID, 0)

            if res_l == vrep.simx_return_ok and res_r == vrep.simx_return_ok and res_p == vrep.simx_return_ok and res_o == vrep.simx_return_ok:
                #print("Converting VREP image to numpy array...")
                cv_image_l = self.VREP_image_to_cv_rgb(
                    self.resolution[1], self.resolution[0], self.image_l)
                cv_image_r = self.VREP_image_to_cv_rgb(
                    self.resolution[1], self.resolution[0], self.image_r)

                self.initalised = True

                return True, cv_image_l, cv_image_r
            else:
                print("Failed to capture vision sensor images: {},{}".format(
                    res_l, res_r))
                return False, None, None
        else:
            print("Failed to connect to VREP")
            return False, None, None

    def get_last_pose(self):
        return self.position, self.orientation

    def close(self):
        vrep.simxStopSimulation(self.clientID, vrep.simx_opmode_oneshot)
        self.disconnect()

    def run(self):
        self.initalised = False
        res = self.connect()
        if (res):
            self.running = True
            while (self.isConnected() and self.running):
                res, cv_image_l, cv_image_r = self.capture()
                if (res):
                    self.cv_image_l = cv_image_l
                    self.cv_image_r = cv_image_r
        self.close()

    def stop_threaded(self):
        self.running = False
        self.close()

    def run_threaded(self):
        self.thread = threading.Thread(target=self.run)
        self.thread.daemon = True
        self.thread.start()


class VREPVisionSensor:
    def __init__(self, name, clientID, en_depth=False):
        self.name = name
        self.clientID = clientID
        self.cv_image = None
        self.cv_dimage = None
        self.running = False
        self.en_depth = en_depth

    def VREP_image_to_cv_rgb(self, h, w, img):
        cv_image = np.array(img, dtype=np.uint8)
        cv_image.resize([h, w, 3])
        cv_image = cv2.flip(cv_image, 0)
        cv_image = cv2.cvtColor(cv_image, cv2.COLOR_BGR2RGB)
        return cv_image

    def VREP_depth_to_np(self, h, w, img):
        cv_image = np.array(img, dtype=np.float32)
        cv_image.resize([h, w])
        cv_image = cv2.flip(cv_image, 0)
        return cv_image

    def disconnect(self):
        # Before closing the connection to V-REP, make sure that the last command sent out had time to arrive. You can guarantee this with (for example):
        vrep.simxGetPingTime(self.clientID)

        # Now close the connection to V-REP:
        vrep.simxFinish(self.clientID)

    def connect(self):
        # print ("Connecting to VREP vision sensor...")
        # vrep.simxFinish(-1) # just in case, close all opened connections
        # self.clientID=vrep.simxStart('127.0.0.1',self.port,True,True,5000,1) # Connect to V-REP
        if self.clientID != -1:
            #print ('Connected to remote API server')
            res, self.handle = vrep.simxGetObjectHandle(
                self.clientID, self.name, vrep.simx_opmode_oneshot_wait)
            if (res == vrep.simx_return_ok):
                res1, self.resolution, self.image = vrep.simxGetVisionSensorImage(
                    self.clientID, self.handle, 0, vrep.simx_opmode_streaming)
                if (self.en_depth):
                    res2, resol, self.depthImage = vrep.simxGetVisionSensorDepthBuffer(
                        self.clientID, self.handle, vrep.simx_opmode_streaming)
                else:
                    res2 = 1
                res3, self.nearClip = vrep.simxGetObjectFloatParameter(
                    self.clientID, self.handle, vrep.sim_visionfloatparam_near_clipping, vrep.simx_opmode_streaming)
                res4, self.farClip = vrep.simxGetObjectFloatParameter(
                    self.clientID, self.handle, vrep.sim_visionfloatparam_far_clipping, vrep.simx_opmode_streaming)
                res5, self.perspective_angle = vrep.simxGetObjectFloatParameter(
                    self.clientID, self.handle, vrep.sim_visionfloatparam_perspective_angle, vrep.simx_opmode_streaming)

                if (res1 == 1 and res2 == 1 and res3 == 1 and res4 == 1 and res5 == 1):
                    return True
                else:
                    print("Failed to set vision sensor properties")
                    print("{},{},{},{},{}".format(
                        res1, res2, res3, res4, res5))
                    return False
            else:
                print("Failed to connect to vision sensor object: {}".format(res))
                return False
        else:
            print("Failed to connect VREP")
            return False

    def isConnected(self):
        if (vrep.simxGetConnectionId(self.clientID) != -1):
            return True
        else:
            return False

    def isInitalised(self):
        return self.initalised

    def capture(self):
        if (vrep.simxGetConnectionId(self.clientID) != -1):
            #print("Getting image from VREP...")
            res1, self.resolution, self.image = vrep.simxGetVisionSensorImage(
                self.clientID, self.handle, 0, vrep.simx_opmode_buffer)
            #res1, self.resolution, self.image = vrep.simxGetVisionSensorImage(self.clientID,self.handle,0,vrep.simx_opmode_oneshot)
            #print("Getting depth image from VREP...")
            if self.en_depth:
                res2, resol, self.depthImage = vrep.simxGetVisionSensorDepthBuffer(
                    self.clientID, self.handle, vrep.simx_opmode_buffer)
            else:
                res2 = vrep.simx_return_ok
            #res2, resol, self.depthImage=vrep.simxGetVisionSensorDepthBuffer(self.clientID,self.handle,vrep.simx_opmode_oneshot)

            if res1 == vrep.simx_return_ok and res2 == vrep.simx_return_ok:
                #print("Converting VREP image to numpy array...")
                cv_image = self.VREP_image_to_cv_rgb(
                    self.resolution[1], self.resolution[0], self.image)
                if self.en_depth:
                    cv_dimage = self.VREP_depth_to_np(
                        self.resolution[1], self.resolution[0], self.depthImage)
                else:
                    cv_dimage = None

                self.initalised = True

                return True, cv_image, cv_dimage
            else:
                #print("Failed to capture vision sensor images: {},{}".format(res1,res2))
                return False, None, None
        else:
            print("Failed to connect to VREP")
            return False, None, None

    def close(self):
        self.disconnect()
        vrep.simxStopSimulation(self.clientID, vrep.simx_opmode_oneshot)
        vrep.simxFinish(self.clientID)

    def run(self):
        self.initalised = False
        res = self.connect()
        if (res):
            self.running = True
            while (self.isConnected() and self.running):
                res, cv_image, cv_dimage = self.capture()
                if (res):
                    self.cv_image = cv_image
                    self.cv_dimage = cv_dimage
        self.close()

    def stop_threaded(self):
        self.running = False
        self.close()

    def run_threaded(self):
        self.thread = threading.Thread(target=self.run)
        self.thread.daemon = True
        self.thread.start()



import cv2
import numpy as np
from pypylon import pylon
from threading import Lock

# Image event handler.
class PylonImageEventHandler(pylon.ImageEventHandler):
    def __init__(self):
        super().__init__()
        self.image_mutex = Lock()
        with self.image_mutex:
            self.image = np.array([])

    def OnImageGrabbed(self, camera, grabResult):
        if grabResult.GrabSucceeded():
            with self.image_mutex:
                self.image = grabResult.GetArray().copy()

    def getImage(self):
        image = np.array([])
        with self.image_mutex:
            image = self.image.copy()
        return image

class PylonCapture():
    PIXEL_FORMAT_MONO8 = "Mono8"

    def __init__(self, serial, trigger_mode=None, pixel_format=None,
                 packet_size=None, inter_packet_delay=None, binning=None,
                 max_buffer=None):
        """
        Initialisation function for PylonCapture class.
        :param serial: serial name of the camera
        :param trigger_mode: camera trigger mode
        :param pixel_format: camera pixel format (e.g. Mono8)
        :param packet_size:
            network packet size. Can be useful to change
            this if you are getting lost packets.
        :param inter_packet_delay:
            network packet delay. Can be useful to change
            this if you are getting lost packets.
        :type serial: string
        :type trigger_mode: bool
        :type pixel_format: string
        :type packet_size: int
        :type inter_packet_delay: int
        """
        self.camera = None
        self.serial = serial
        self.trigger_mode = trigger_mode
        self.pixel_format = pixel_format
        self.packet_size = packet_size
        self.inter_packet_delay = inter_packet_delay
        self.binning = binning
        self.max_buffer = max_buffer

    def connect(self):
        """
        Connect to basler camera using Pylon interface.
        :returns: success
        :rtype: bool
        """
        # Get the transport layer factory.
        tlFactory = pylon.TlFactory.GetInstance()

        device = None

        for d in tlFactory.GetInstance().EnumerateDevices():
            if d.GetSerialNumber() == self.serial:
                device = d
                break
        else:
            print('Camera with {} serial number not found'.format(self.serial))
            return False

        self.camera = pylon.InstantCamera(
            pylon.TlFactory.GetInstance().CreateDevice(device))

        self.converter = pylon.ImageFormatConverter()
        # converter.OutputPixelFormat = pylon.PixelType_BGR8packed
        self.converter.OutputPixelFormat = pylon.PixelType_Mono8
        self.converter.OutputBitAlignment = pylon.OutputBitAlignment_MsbAligned

        self.camera.RegisterConfiguration(pylon.ConfigurationEventHandler(), pylon.RegistrationMode_ReplaceAll,
                                        pylon.Cleanup_Delete)
        self.image_event_handler = PylonImageEventHandler()
        self.camera.RegisterImageEventHandler(self.image_event_handler, pylon.RegistrationMode_Append, pylon.Cleanup_Delete)

        # TODO fix pypylon problem where settings are
        # reset on startup (previous settings are wiped)
        self.camera.Open()

        if self.trigger_mode is not None:
            print("Trigger mode: {}".format(self.trigger_mode))
            if self.trigger_mode:
                self.camera.GetNodeMap().GetNode("TriggerMode").SetValue("On")
            else:
                self.camera.GetNodeMap().GetNode("TriggerMode").SetValue("Off")
        if self.max_buffer is not None:
            print("Max buffer: {}".format(self.max_buffer))
            self.camera.MaxNumBuffer.SetValue(self.max_buffer)
        # TODO set exposure and gain in functions
        # self.camera.GetNodeMap().GetNode("ExposureAuto").SetValue("Off")
        # self.camera.GetNodeMap().GetNode("ExposureTimeRaw").SetValue(15000)
        # self.camera.GetNodeMap().GetNode("GainAuto").SetValue("Off")
        # self.camera.GetNodeMap().GetNode("GainRaw").SetValue(0)
        if self.pixel_format is not None:
            print("Pixel format: {}".format(self.pixel_format))
            self.camera.GetNodeMap().GetNode("PixelFormat").SetValue(self.pixel_format)
        # self.camera.GetNodeMap().GetNode("Width").SetValue(2448)
        # self.camera.GetNodeMap().GetNode("Height").SetValue(2048)
        if self.packet_size is not None:
            print("Packet size: {}".format(self.packet_size))
            self.camera.GetNodeMap().GetNode("GevSCPSPacketSize").SetValue(self.packet_size)
        if self.inter_packet_delay is not None:
            print("Inter packet delay: {}".format(self.inter_packet_delay))
            self.camera.GetNodeMap().GetNode("GevSCPD").SetValue(self.inter_packet_delay)
        if self.binning is not None:
            print("Binning: {}".format(self.binning))
            self.camera.GetNodeMap().GetNode("BinningHorizontalMode").SetValue("Average")
            self.camera.GetNodeMap().GetNode("BinningHorizontal").SetValue(self.binning)
            self.camera.GetNodeMap().GetNode("BinningVerticalMode").SetValue("Average")
            self.camera.GetNodeMap().GetNode("BinningVertical").SetValue(self.binning)

        self.camera.StartGrabbing(pylon.GrabStrategy_LatestImageOnly, pylon.GrabLoop_ProvidedByInstantCamera)

        return True

    def grab(self):
        """
        Grab image from camera
        :returns: success of capture, image
        :rtype: bool, numpy.array
        """
        if self.camera is not None:
            if self.camera.IsGrabbing():
                frame = self.image_event_handler.getImage()
                if frame.size != 0:
                    return True, frame
                else:
                    return False, None
            else:
                return False, None
        else:
            return False, None

    def close(self):
        """
        Close connection to camera
        """
        self.camera.StopGrabbing()
        self.camera.Close()


if __name__ == '__main__':
    CAMERA_SERIAL = "23517286"
    cam = PylonCapture(CAMERA_SERIAL)
    cam.connect()
    while True:
        res, image = cam.grab()
        if res:
            image_resized = cv2.resize(image, (640, 480))
            cv2.imshow('Image', image_resized)
            k = cv2.waitKey(1)
            if k == ord('q'):
                break
    cam.close()


# constants
# Scene object types. Values are serialized
sim_object_shape_type = 0
sim_object_joint_type = 1
sim_object_graph_type = 2
sim_object_camera_type = 3
sim_object_dummy_type = 4
sim_object_proximitysensor_type = 5
sim_object_reserved1 = 6
sim_object_reserved2 = 7
sim_object_path_type = 8
sim_object_visionsensor_type = 9
sim_object_volume_type = 10
sim_object_mill_type = 11
sim_object_forcesensor_type = 12
sim_object_light_type = 13
sim_object_mirror_type = 14

# General object types. Values are serialized
sim_appobj_object_type = 109
sim_appobj_collision_type = 110
sim_appobj_distance_type = 111
sim_appobj_simulation_type = 112
sim_appobj_ik_type = 113
sim_appobj_constraintsolver_type = 114
sim_appobj_collection_type = 115
sim_appobj_ui_type = 116
sim_appobj_script_type = 117
sim_appobj_pathplanning_type = 118
sim_appobj_RESERVED_type = 119
sim_appobj_texture_type = 120

# Ik calculation methods. Values are serialized
sim_ik_pseudo_inverse_method = 0
sim_ik_damped_least_squares_method = 1
sim_ik_jacobian_transpose_method = 2

# Ik constraints. Values are serialized
sim_ik_x_constraint = 1
sim_ik_y_constraint = 2
sim_ik_z_constraint = 4
sim_ik_alpha_beta_constraint = 8
sim_ik_gamma_constraint = 16
sim_ik_avoidance_constraint = 64

# Ik calculation results
sim_ikresult_not_performed = 0
sim_ikresult_success = 1
sim_ikresult_fail = 2

# Scene object sub-types. Values are serialized
# Light sub-types
sim_light_omnidirectional_subtype = 1
sim_light_spot_subtype = 2
sim_light_directional_subtype = 3
# Joint sub-types
sim_joint_revolute_subtype = 10
sim_joint_prismatic_subtype = 11
sim_joint_spherical_subtype = 12
# Shape sub-types
sim_shape_simpleshape_subtype = 20
sim_shape_multishape_subtype = 21
# Proximity sensor sub-types
sim_proximitysensor_pyramid_subtype = 30
sim_proximitysensor_cylinder_subtype = 31
sim_proximitysensor_disc_subtype = 32
sim_proximitysensor_cone_subtype = 33
sim_proximitysensor_ray_subtype = 34
# Mill sub-types
sim_mill_pyramid_subtype = 40
sim_mill_cylinder_subtype = 41
sim_mill_disc_subtype = 42
sim_mill_cone_subtype = 42
# No sub-type
sim_object_no_subtype = 200


# Scene object main properties (serialized)
sim_objectspecialproperty_collidable = 0x0001
sim_objectspecialproperty_measurable = 0x0002
#reserved                        =0x0004
#reserved                        =0x0008
sim_objectspecialproperty_detectable_ultrasonic = 0x0010
sim_objectspecialproperty_detectable_infrared = 0x0020
sim_objectspecialproperty_detectable_laser = 0x0040
sim_objectspecialproperty_detectable_inductive = 0x0080
sim_objectspecialproperty_detectable_capacitive = 0x0100
sim_objectspecialproperty_renderable = 0x0200
sim_objectspecialproperty_detectable_all = sim_objectspecialproperty_detectable_ultrasonic | sim_objectspecialproperty_detectable_infrared | sim_objectspecialproperty_detectable_laser | sim_objectspecialproperty_detectable_inductive | sim_objectspecialproperty_detectable_capacitive
sim_objectspecialproperty_cuttable = 0x0400
sim_objectspecialproperty_pathplanning_ignored = 0x0800

# Model properties (serialized)
sim_modelproperty_not_collidable = 0x0001
sim_modelproperty_not_measurable = 0x0002
sim_modelproperty_not_renderable = 0x0004
sim_modelproperty_not_detectable = 0x0008
sim_modelproperty_not_cuttable = 0x0010
sim_modelproperty_not_dynamic = 0x0020
# cannot be selected if sim_modelproperty_not_dynamic is not selected
sim_modelproperty_not_respondable = 0x0040
# Model is not reset at simulation end. This flag is cleared at simulation end
sim_modelproperty_not_reset = 0x0080
# Whole model is invisible independent of local visibility settings
sim_modelproperty_not_visible = 0x0100
sim_modelproperty_not_model = 0xf000  # object is not a model


# Check the documentation instead of comments below!!
# Following messages are dispatched to the Lua-message container
# a UI button slider etc. changed (due to a user's action). aux[0]=UI handle aux[1]=button handle aux[2]=button attributes aux[3]=slider position (if slider)
sim_message_ui_button_state_change = 0
sim_message_reserved9 = 1    # Do not use
sim_message_object_selection_changed = 2
sim_message_reserved10 = 3    # do not use
sim_message_model_loaded = 4
sim_message_reserved11 = 5    # do not use
# a key was pressed while the focus was on a page (aux[0]=key aux[1]=ctrl and shift key state)
sim_message_keypress = 6
sim_message_bannerclicked = 7    # a banner was clicked (aux[0]=banner ID)


# Following messages are dispatched only to the C-API (not available from Lua)
sim_message_for_c_api_only_start = 0x100      # Do not use
sim_message_reserved1 = 0x101      # Do not use
sim_message_reserved2 = 0x102      # Do not use
sim_message_reserved3 = 0x103      # Do not use
sim_message_eventcallback_scenesave = 0x104        # about to save a scene
# about to save a model (current selection will be saved)
sim_message_eventcallback_modelsave = 0x105
# called when simOpenModule in Lua is called
sim_message_eventcallback_moduleopen = 0x106
# called when simHandleModule in Lua is called with argument false
sim_message_eventcallback_modulehandle = 0x107
# called when simCloseModule in Lua is called
sim_message_eventcallback_moduleclose = 0x108
sim_message_reserved4 = 0x109      # Do not use
sim_message_reserved5 = 0x10a        # Do not use
sim_message_reserved6 = 0x10b        # Do not use
sim_message_reserved7 = 0x10c        # Do not use
# Called once every main application loop pass. auxiliaryData[0] contains event flags of events that happened since last time
sim_message_eventcallback_instancepass = 0x10d
sim_message_eventcallback_broadcast = 0x10e
sim_message_eventcallback_imagefilter_enumreset = 0x10f
sim_message_eventcallback_imagefilter_enumerate = 0x110
sim_message_eventcallback_imagefilter_adjustparams = 0x111
sim_message_eventcallback_imagefilter_reserved = 0x112
sim_message_eventcallback_imagefilter_process = 0x113
sim_message_eventcallback_reserved1 = 0x114   # do not use
sim_message_eventcallback_reserved2 = 0x115   # do not use
sim_message_eventcallback_reserved3 = 0x116   # do not use
sim_message_eventcallback_reserved4 = 0x117   # do not use
# the undo button was hit and a previous state is about to be restored
sim_message_eventcallback_abouttoundo = 0x118
# the undo button was hit and a previous state restored
sim_message_eventcallback_undoperformed = 0x119
# the redo button was hit and a future state is about to be restored
sim_message_eventcallback_abouttoredo = 0x11a
# the redo button was hit and a future state restored
sim_message_eventcallback_redoperformed = 0x11b
# scipt icon was double clicked.  (aux[0]=object handle associated with script set replyData[0] to 1 if script should not be opened)
sim_message_eventcallback_scripticondblclick = 0x11c
sim_message_eventcallback_simulationabouttostart = 0x11d
sim_message_eventcallback_simulationended = 0x11e
sim_message_eventcallback_reserved5 = 0x11f   # do not use
# a key was pressed while the focus was on a page (aux[0]=key aux[1]=ctrl and shift key state)
sim_message_eventcallback_keypress = 0x120
# called when simHandleModule in Lua is called with argument true
sim_message_eventcallback_modulehandleinsensingpart = 0x121
# called just before the scene is rendered
sim_message_eventcallback_renderingpass = 0x122
# called when a banner was clicked (aux[0]=banner ID)
sim_message_eventcallback_bannerclicked = 0x123
# auxiliaryData[0] indicates the handle of the item auxiliaryData[1] indicates the state of the item
sim_message_eventcallback_menuitemselected = 0x124
# aux[0]=refresh degree (0=light 1=medium 2=full)
sim_message_eventcallback_refreshdialogs = 0x125
sim_message_eventcallback_sceneloaded = 0x126
sim_message_eventcallback_modelloaded = 0x127
sim_message_eventcallback_instanceswitch = 0x128
sim_message_eventcallback_guipass = 0x129
sim_message_eventcallback_mainscriptabouttobecalled = 0x12a
# the command simRMLPosition was called. The appropriate plugin should handle the call
sim_message_eventcallback_rmlposition = 0x12b
# the command simRMLVelocity was called. The appropriate plugin should handle the call
sim_message_eventcallback_rmlvelocity = 0x12c
sim_message_simulation_start_resume_request = 0x1000
sim_message_simulation_pause_request = 0x1001
sim_message_simulation_stop_request = 0x1002

# Scene object properties. Combine with the | operator
sim_objectproperty_reserved1 = 0x0000
sim_objectproperty_reserved2 = 0x0001
sim_objectproperty_reserved3 = 0x0002
sim_objectproperty_reserved4 = 0x0003
sim_objectproperty_reserved5 = 0x0004  # formely sim_objectproperty_visible
sim_objectproperty_reserved6 = 0x0008  # formely sim_objectproperty_wireframe
sim_objectproperty_collapsed = 0x0010
sim_objectproperty_selectable = 0x0020
sim_objectproperty_reserved7 = 0x0040
sim_objectproperty_selectmodelbaseinstead = 0x0080
sim_objectproperty_dontshowasinsidemodel = 0x0100
# reserved                                    =0x0200
sim_objectproperty_canupdatedna = 0x0400
sim_objectproperty_selectinvisible = 0x0800
sim_objectproperty_depthinvisible = 0x1000


# type of arguments (input and output) for custom lua commands
sim_lua_arg_nil = 0
sim_lua_arg_bool = 1
sim_lua_arg_int = 2
sim_lua_arg_float = 3
sim_lua_arg_string = 4
sim_lua_arg_invalid = 5
sim_lua_arg_table = 8

# custom user interface properties. Values are serialized.
sim_ui_property_visible = 0x0001
sim_ui_property_visibleduringsimulationonly = 0x0002
sim_ui_property_moveable = 0x0004
sim_ui_property_relativetoleftborder = 0x0008
sim_ui_property_relativetotopborder = 0x0010
sim_ui_property_fixedwidthfont = 0x0020
sim_ui_property_systemblock = 0x0040
sim_ui_property_settocenter = 0x0080
sim_ui_property_rolledup = 0x0100
sim_ui_property_selectassociatedobject = 0x0200
sim_ui_property_visiblewhenobjectselected = 0x0400


# button properties. Values are serialized.
sim_buttonproperty_button = 0x0000
sim_buttonproperty_label = 0x0001
sim_buttonproperty_slider = 0x0002
sim_buttonproperty_editbox = 0x0003
sim_buttonproperty_staydown = 0x0008
sim_buttonproperty_enabled = 0x0010
sim_buttonproperty_borderless = 0x0020
sim_buttonproperty_horizontallycentered = 0x0040
sim_buttonproperty_ignoremouse = 0x0080
sim_buttonproperty_isdown = 0x0100
sim_buttonproperty_transparent = 0x0200
sim_buttonproperty_nobackgroundcolor = 0x0400
sim_buttonproperty_rollupaction = 0x0800
sim_buttonproperty_closeaction = 0x1000
sim_buttonproperty_verticallycentered = 0x2000
sim_buttonproperty_downupevent = 0x4000


# Simulation status
# Simulation is stopped
sim_simulation_stopped = 0x00
# Simulation is paused
sim_simulation_paused = 0x08
# Simulation is advancing
sim_simulation_advancing = 0x10
# First simulation pass (1x)
sim_simulation_advancing_firstafterstop = sim_simulation_advancing | 0x00
# Normal simulation pass (>=1x)
sim_simulation_advancing_running = sim_simulation_advancing | 0x01
# reserved                                    =sim_simulation_advancing|0x02
# Last simulation pass before pause (1x)
sim_simulation_advancing_lastbeforepause = sim_simulation_advancing | 0x03
# First simulation pass after pause (1x)
sim_simulation_advancing_firstafterpause = sim_simulation_advancing | 0x04
# "Trying to stop" simulation pass (>=1x)
sim_simulation_advancing_abouttostop = sim_simulation_advancing | 0x05
# Last simulation pass (1x)
sim_simulation_advancing_lastbeforestop = sim_simulation_advancing | 0x06


# Script execution result (first return value)
sim_script_no_error = 0
sim_script_main_script_nonexistent = 1
sim_script_main_script_not_called = 2
sim_script_reentrance_error = 4
sim_script_lua_error = 8
sim_script_call_error = 16

# Script types (serialized!)
sim_scripttype_mainscript = 0
sim_scripttype_childscript = 1
sim_scripttype_jointctrlcallback = 4
sim_scripttype_contactcallback = 5
sim_scripttype_customizationscript = 6
sim_scripttype_generalcallback = 7

# API call error messages
sim_api_errormessage_ignore = 0    # does not memorize nor output errors
sim_api_errormessage_report = 1    # memorizes errors (default for C-API calls)
# memorizes and outputs errors (default for Lua-API calls)
sim_api_errormessage_output = 2


# special argument of some functions
sim_handle_all = -2
sim_handle_all_except_explicit = -3
sim_handle_self = -4
sim_handle_main_script = -5
sim_handle_tree = -6
sim_handle_chain = -7
sim_handle_single = -8
sim_handle_default = -9
sim_handle_all_except_self = -10
sim_handle_parent = -11


# special handle flags
sim_handleflag_assembly = 0x400000
sim_handleflag_model = 0x800000


# distance calculation methods (serialized)
sim_distcalcmethod_dl = 0
sim_distcalcmethod_dac = 1
sim_distcalcmethod_max_dl_dac = 2
sim_distcalcmethod_dl_and_dac = 3
sim_distcalcmethod_sqrt_dl2_and_dac2 = 4
sim_distcalcmethod_dl_if_nonzero = 5
sim_distcalcmethod_dac_if_nonzero = 6

# Generic dialog styles
sim_dlgstyle_message = 0
sim_dlgstyle_input = 1
sim_dlgstyle_ok = 2
sim_dlgstyle_ok_cancel = 3
sim_dlgstyle_yes_no = 4
# can be combined with one of above values. Only with this flag can the position of the related UI be set just after dialog creation
sim_dlgstyle_dont_center = 32

# Generic dialog return values
sim_dlgret_still_open = 0
sim_dlgret_ok = 1
sim_dlgret_cancel = 2
sim_dlgret_yes = 3
sim_dlgret_no = 4


# Path properties
sim_pathproperty_show_line = 0x0001
sim_pathproperty_show_orientation = 0x0002
sim_pathproperty_closed_path = 0x0004
sim_pathproperty_automatic_orientation = 0x0008
sim_pathproperty_invert_velocity = 0x0010
sim_pathproperty_infinite_acceleration = 0x0020
sim_pathproperty_flat_path = 0x0040
sim_pathproperty_show_position = 0x0080
sim_pathproperty_auto_velocity_profile_translation = 0x0100
sim_pathproperty_auto_velocity_profile_rotation = 0x0200
sim_pathproperty_endpoints_at_zero = 0x0400
sim_pathproperty_keep_x_up = 0x0800

# drawing objects
# following are mutually exclusive
sim_drawing_points = 0            # 3 values per point (point size in pixels)
sim_drawing_lines = 1            # 6 values per line (line size in pixels)
sim_drawing_triangles = 2            # 9 values per triangle
# 6 values per point (3 for triangle position 3 for triangle normal vector) (triangle size in meters)
sim_drawing_trianglepoints = 3
# 6 values per point (3 for quad position 3 for quad normal vector) (quad size in meters)
sim_drawing_quadpoints = 4
# 6 values per point (3 for disc position 3 for disc normal vector) (disc size in meters)
sim_drawing_discpoints = 5
# 6 values per point (3 for cube position 3 for cube normal vector) (cube size in meters)
sim_drawing_cubepoints = 6
# 3 values per point (sphere size in meters)
sim_drawing_spherepoints = 7

# following can be or-combined
# +3 values per item (each item has its own ambient color (rgb values)).
sim_drawing_itemcolors = 0x00020
# Mutually exclusive with sim_drawing_vertexcolors
# +3 values per vertex (each vertex has its own ambient color (rgb values). Only for sim_drawing_lines (+6) and for sim_drawing_triangles(+9)). Mutually exclusive with sim_drawing_itemcolors
sim_drawing_vertexcolors = 0x00040
# +1 value per item (each item has its own size). Not for sim_drawing_triangles
sim_drawing_itemsizes = 0x00080
# back faces are not displayed for all items
sim_drawing_backfaceculling = 0x00100
sim_drawing_wireframe = 0x00200  # all items displayed in wireframe
# all items are tagged as paint (for additinal processing at a later stage)
sim_drawing_painttag = 0x00400
# if the object is associated with a scene object then it follows that visibility otherwise it is always visible
sim_drawing_followparentvisibility = 0x00800
# if the max item count was reached then the first items are overwritten.
sim_drawing_cyclic = 0x01000
# the drawing object will be 50% transparent
sim_drawing_50percenttransparency = 0x02000
# the drawing object will be 25% transparent
sim_drawing_25percenttransparency = 0x04000
# the drawing object will be 12.5% transparent
sim_drawing_12percenttransparency = 0x08000
# When used in combination with sim_drawing_itemcolors or sim_drawing_vertexcolors then the specified colors will be for the emissive component
sim_drawing_emissioncolor = 0x10000
# Only for trianglepoints quadpoints discpoints and cubepoints. If specified the normal verctor is calculated to face the camera (each item data requires 3 values less)
sim_drawing_facingcamera = 0x20000
# When specified objects are always drawn on top of "regular objects"
sim_drawing_overlay = 0x40000
# +1 value per item (each item has its own transparency value (0-1)). Not compatible with sim_drawing_vertexcolors
sim_drawing_itemtransparency = 0x80000

# banner values
# following can be or-combined
sim_banner_left = 0x00001  # Banners display on the left of the specified point
sim_banner_right = 0x00002  # Banners display on the right of the specified point
sim_banner_nobackground = 0x00004  # Banners have no background rectangle
# When specified banners are always drawn on top of "regular objects"
sim_banner_overlay = 0x00008
# if the object is associated with a scene object then it follows that visibility otherwise it is always visible
sim_banner_followparentvisibility = 0x00010
# if the object is associated with a scene object then clicking the banner will select the scene object
sim_banner_clickselectsparent = 0x00020
# if the banner is clicked an event is triggered (sim_message_eventcallback_bannerclicked and sim_message_bannerclicked are generated)
sim_banner_clicktriggersevent = 0x00040
# If specified the banner will always face the camera by rotating around the banner's vertical axis (y-axis)
sim_banner_facingcamera = 0x00080
# If specified the banner will always fully face the camera (the banner's orientation is same as the camera looking at it)
sim_banner_fullyfacingcamera = 0x00100
# If specified the banner will only be visible from one side
sim_banner_backfaceculling = 0x00200
# If specified the banner will always appear in the same size. In that case size represents the character height in pixels
sim_banner_keepsamesize = 0x00400
# If specified a fixed-size bitmap font is used. The text will also always fully face the camera and be right
sim_banner_bitmapfont = 0x00800
# to the specified position. Bitmap fonts are not clickable


# particle objects following are mutually exclusive
# 6 values per point (pt1 and pt2. Pt1 is start position pt2-pt1 is the initial velocity vector). i
sim_particle_points1 = 0
# Point is 1 pixel big. Only appearance is a point internally handled as a perfect sphere
# 6 values per point. Point is 2 pixel big. Only appearance is a point internally handled as a perfect sphere
sim_particle_points2 = 1
# 6 values per point. Point is 4 pixel big. Only appearance is a point internally handled as a perfect sphere
sim_particle_points4 = 2
# 6 values per sphere. Only appearance is rough. Internally a perfect sphere
sim_particle_roughspheres = 3
sim_particle_spheres = 4    # 6 values per sphere. Internally a perfect sphere


# following can be or-combined
# the particles are respondable against shapes (against all objects that have at least one bit 1-4 activated in the global respondable mask)
sim_particle_respondable1to4 = 0x0020
# the particles are respondable against shapes (against all objects that have at least one bit 5-8 activated in the global respondable mask)
sim_particle_respondable5to8 = 0x0040
# the particles are respondable against each other
sim_particle_particlerespondable = 0x0080
# the particles ignore the effect of gravity. Not compatible with sim_particle_water
sim_particle_ignoresgravity = 0x0100
sim_particle_invisible = 0x0200  # the particles are invisible
# +1 value per particle (each particle can have a different size)
sim_particle_itemsizes = 0x0400
# +1 value per particle (each particle can have a different density)
sim_particle_itemdensities = 0x0800
# +3 values per particle (each particle can have a different color)
sim_particle_itemcolors = 0x1000
# if the max item count was reached then the first items are overwritten.
sim_particle_cyclic = 0x2000
# When used in combination with sim_particle_itemcolors then the specified colors will be for the emissive component
sim_particle_emissioncolor = 0x4000
# the particles are water particles (no weight in the water (i.e. when z<0)). Not compatible with sim_particle_ignoresgravity
sim_particle_water = 0x8000
# The particles can be seen by vision sensors (sim_particle_invisible must not be set)
sim_particle_painttag = 0x10000


# custom user interface menu attributes
sim_ui_menu_title = 1
sim_ui_menu_minimize = 2
sim_ui_menu_close = 4
sim_ui_menu_systemblock = 8


# Boolean parameters
sim_boolparam_hierarchy_visible = 0
sim_boolparam_console_visible = 1
sim_boolparam_collision_handling_enabled = 2
sim_boolparam_distance_handling_enabled = 3
sim_boolparam_ik_handling_enabled = 4
sim_boolparam_gcs_handling_enabled = 5
sim_boolparam_dynamics_handling_enabled = 6
sim_boolparam_joint_motion_handling_enabled = 7
sim_boolparam_path_motion_handling_enabled = 8
sim_boolparam_proximity_sensor_handling_enabled = 9
sim_boolparam_vision_sensor_handling_enabled = 10
sim_boolparam_mill_handling_enabled = 11
sim_boolparam_browser_visible = 12
sim_boolparam_scene_and_model_load_messages = 13
sim_reserved0 = 14
sim_boolparam_shape_textures_are_visible = 15
sim_boolparam_display_enabled = 16
sim_boolparam_infotext_visible = 17
sim_boolparam_statustext_open = 18
sim_boolparam_fog_enabled = 19
sim_boolparam_rml2_available = 20
sim_boolparam_rml4_available = 21
sim_boolparam_mirrors_enabled = 22
sim_boolparam_aux_clip_planes_enabled = 23
sim_boolparam_full_model_copy_from_api = 24
sim_boolparam_realtime_simulation = 25
sim_boolparam_force_show_wireless_emission = 27
sim_boolparam_force_show_wireless_reception = 28
sim_boolparam_video_recording_triggered = 29
sim_boolparam_threaded_rendering_enabled = 32
sim_boolparam_fullscreen = 33
sim_boolparam_headless = 34
sim_boolparam_hierarchy_toolbarbutton_enabled = 35
sim_boolparam_browser_toolbarbutton_enabled = 36
sim_boolparam_objectshift_toolbarbutton_enabled = 37
sim_boolparam_objectrotate_toolbarbutton_enabled = 38
sim_boolparam_force_calcstruct_all_visible = 39
sim_boolparam_force_calcstruct_all = 40
sim_boolparam_exit_request = 41
sim_boolparam_play_toolbarbutton_enabled = 42
sim_boolparam_pause_toolbarbutton_enabled = 43
sim_boolparam_stop_toolbarbutton_enabled = 44
sim_boolparam_waiting_for_trigger = 45


# Integer parameters
# Check sim_api_errormessage_... constants above for valid values
sim_intparam_error_report_mode = 0
sim_intparam_program_version = 1  # e.g Version 2.1.4 --> 20104. Can only be read
# do not use anymore (always returns 1 since CoppeliaSim 2.5.11)
sim_intparam_instance_count = 2
sim_intparam_custom_cmd_start_id = 3  # can only be read
# 0=evaluation version 1=full version 2=player version. Can only be read
sim_intparam_compilation_version = 4
sim_intparam_current_page = 5
sim_intparam_flymode_camera_handle = 6  # can only be read
sim_intparam_dynamic_step_divider = 7  # can only be read
sim_intparam_dynamic_engine = 8  # 0=Bullet 1=ODE. 2=Vortex.
sim_intparam_server_port_start = 9  # can only be read
sim_intparam_server_port_range = 10  # can only be read
sim_intparam_visible_layers = 11
sim_intparam_infotext_style = 12
sim_intparam_settings = 13
sim_intparam_edit_mode_type = 14  # can only be read
# is initialized at sim_intparam_server_port_start
sim_intparam_server_port_next = 15
sim_intparam_qt_version = 16  # version of the used Qt framework
sim_intparam_event_flags_read = 17  # can only be read
sim_intparam_event_flags_read_clear = 18  # can only be read
sim_intparam_platform = 19  # can only be read
sim_intparam_scene_unique_id = 20  # can only be read
sim_intparam_work_thread_count = 21
sim_intparam_mouse_x = 22
sim_intparam_mouse_y = 23
sim_intparam_core_count = 24
sim_intparam_work_thread_calc_time_ms = 25
sim_intparam_idle_fps = 26
sim_intparam_prox_sensor_select_down = 27
sim_intparam_prox_sensor_select_up = 28
sim_intparam_stop_request_counter = 29
sim_intparam_program_revision = 30
sim_intparam_mouse_buttons = 31
sim_intparam_dynamic_warning_disabled_mask = 32
sim_intparam_simulation_warning_disabled_mask = 33
sim_intparam_scene_index = 34
sim_intparam_motionplanning_seed = 35
sim_intparam_speedmodifier = 36

# Float parameters
sim_floatparam_rand = 0  # random value (0.0-1.0)
sim_floatparam_simulation_time_step = 1
sim_floatparam_stereo_distance = 2

# String parameters
sim_stringparam_application_path = 0  # path of CoppeliaSim's executable
sim_stringparam_video_filename = 1
sim_stringparam_app_arg1 = 2
sim_stringparam_app_arg2 = 3
sim_stringparam_app_arg3 = 4
sim_stringparam_app_arg4 = 5
sim_stringparam_app_arg5 = 6
sim_stringparam_app_arg6 = 7
sim_stringparam_app_arg7 = 8
sim_stringparam_app_arg8 = 9
sim_stringparam_app_arg9 = 10
sim_stringparam_scene_path_and_name = 13

# Array parameters
sim_arrayparam_gravity = 0
sim_arrayparam_fog = 1
sim_arrayparam_fog_color = 2
sim_arrayparam_background_color1 = 3
sim_arrayparam_background_color2 = 4
sim_arrayparam_ambient_light = 5
sim_arrayparam_random_euler = 6

sim_objintparam_visibility_layer = 10
sim_objfloatparam_abs_x_velocity = 11
sim_objfloatparam_abs_y_velocity = 12
sim_objfloatparam_abs_z_velocity = 13
sim_objfloatparam_abs_rot_velocity = 14
sim_objfloatparam_objbbox_min_x = 15
sim_objfloatparam_objbbox_min_y = 16
sim_objfloatparam_objbbox_min_z = 17
sim_objfloatparam_objbbox_max_x = 18
sim_objfloatparam_objbbox_max_y = 19
sim_objfloatparam_objbbox_max_z = 20
sim_objfloatparam_modelbbox_min_x = 21
sim_objfloatparam_modelbbox_min_y = 22
sim_objfloatparam_modelbbox_min_z = 23
sim_objfloatparam_modelbbox_max_x = 24
sim_objfloatparam_modelbbox_max_y = 25
sim_objfloatparam_modelbbox_max_z = 26
sim_objintparam_collection_self_collision_indicator = 27
sim_objfloatparam_transparency_offset = 28
sim_objintparam_child_role = 29
sim_objintparam_parent_role = 30
sim_objintparam_manipulation_permissions = 31
sim_objintparam_illumination_handle = 32

sim_visionfloatparam_near_clipping = 1000
sim_visionfloatparam_far_clipping = 1001
sim_visionintparam_resolution_x = 1002
sim_visionintparam_resolution_y = 1003
sim_visionfloatparam_perspective_angle = 1004
sim_visionfloatparam_ortho_size = 1005
sim_visionintparam_disabled_light_components = 1006
sim_visionintparam_rendering_attributes = 1007
sim_visionintparam_entity_to_render = 1008
sim_visionintparam_windowed_size_x = 1009
sim_visionintparam_windowed_size_y = 1010
sim_visionintparam_windowed_pos_x = 1011
sim_visionintparam_windowed_pos_y = 1012
sim_visionintparam_pov_focal_blur = 1013
sim_visionfloatparam_pov_blur_distance = 1014
sim_visionfloatparam_pov_aperture = 1015
sim_visionintparam_pov_blur_sampled = 1016
sim_visionintparam_render_mode = 1017

sim_jointintparam_motor_enabled = 2000
sim_jointintparam_ctrl_enabled = 2001
sim_jointfloatparam_pid_p = 2002
sim_jointfloatparam_pid_i = 2003
sim_jointfloatparam_pid_d = 2004
sim_jointfloatparam_intrinsic_x = 2005
sim_jointfloatparam_intrinsic_y = 2006
sim_jointfloatparam_intrinsic_z = 2007
sim_jointfloatparam_intrinsic_qx = 2008
sim_jointfloatparam_intrinsic_qy = 2009
sim_jointfloatparam_intrinsic_qz = 2010
sim_jointfloatparam_intrinsic_qw = 2011
sim_jointfloatparam_velocity = 2012
sim_jointfloatparam_spherical_qx = 2013
sim_jointfloatparam_spherical_qy = 2014
sim_jointfloatparam_spherical_qz = 2015
sim_jointfloatparam_spherical_qw = 2016
sim_jointfloatparam_upper_limit = 2017
sim_jointfloatparam_kc_k = 2018
sim_jointfloatparam_kc_c = 2019
sim_jointfloatparam_ik_weight = 2021
sim_jointfloatparam_error_x = 2022
sim_jointfloatparam_error_y = 2023
sim_jointfloatparam_error_z = 2024
sim_jointfloatparam_error_a = 2025
sim_jointfloatparam_error_b = 2026
sim_jointfloatparam_error_g = 2027
sim_jointfloatparam_error_pos = 2028
sim_jointfloatparam_error_angle = 2029
sim_jointintparam_velocity_lock = 2030
sim_jointintparam_vortex_dep_handle = 2031
sim_jointfloatparam_vortex_dep_multiplication = 2032
sim_jointfloatparam_vortex_dep_offset = 2033

sim_shapefloatparam_init_velocity_x = 3000
sim_shapefloatparam_init_velocity_y = 3001
sim_shapefloatparam_init_velocity_z = 3002
sim_shapeintparam_static = 3003
sim_shapeintparam_respondable = 3004
sim_shapefloatparam_mass = 3005
sim_shapefloatparam_texture_x = 3006
sim_shapefloatparam_texture_y = 3007
sim_shapefloatparam_texture_z = 3008
sim_shapefloatparam_texture_a = 3009
sim_shapefloatparam_texture_b = 3010
sim_shapefloatparam_texture_g = 3011
sim_shapefloatparam_texture_scaling_x = 3012
sim_shapefloatparam_texture_scaling_y = 3013
sim_shapeintparam_culling = 3014
sim_shapeintparam_wireframe = 3015
sim_shapeintparam_compound = 3016
sim_shapeintparam_convex = 3017
sim_shapeintparam_convex_check = 3018
sim_shapeintparam_respondable_mask = 3019
sim_shapefloatparam_init_velocity_a = 3020
sim_shapefloatparam_init_velocity_b = 3021
sim_shapefloatparam_init_velocity_g = 3022
sim_shapestringparam_color_name = 3023
sim_shapeintparam_edge_visibility = 3024
sim_shapefloatparam_shading_angle = 3025
sim_shapefloatparam_edge_angle = 3026
sim_shapeintparam_edge_borders_hidden = 3027

sim_proxintparam_ray_invisibility = 4000

sim_forcefloatparam_error_x = 5000
sim_forcefloatparam_error_y = 5001
sim_forcefloatparam_error_z = 5002
sim_forcefloatparam_error_a = 5003
sim_forcefloatparam_error_b = 5004
sim_forcefloatparam_error_g = 5005
sim_forcefloatparam_error_pos = 5006
sim_forcefloatparam_error_angle = 5007

sim_lightintparam_pov_casts_shadows = 8000

sim_cameraintparam_disabled_light_components = 9000
sim_camerafloatparam_perspective_angle = 9001
sim_camerafloatparam_ortho_size = 9002
sim_cameraintparam_rendering_attributes = 9003
sim_cameraintparam_pov_focal_blur = 9004
sim_camerafloatparam_pov_blur_distance = 9005
sim_camerafloatparam_pov_aperture = 9006
sim_cameraintparam_pov_blur_samples = 9007

sim_dummyintparam_link_type = 10000

sim_mirrorfloatparam_width = 12000
sim_mirrorfloatparam_height = 12001
sim_mirrorfloatparam_reflectance = 12002
sim_mirrorintparam_enable = 12003

sim_pplanfloatparam_x_min = 20000
sim_pplanfloatparam_x_range = 20001
sim_pplanfloatparam_y_min = 20002
sim_pplanfloatparam_y_range = 20003
sim_pplanfloatparam_z_min = 20004
sim_pplanfloatparam_z_range = 20005
sim_pplanfloatparam_delta_min = 20006
sim_pplanfloatparam_delta_range = 20007

sim_mplanintparam_nodes_computed = 25000
sim_mplanintparam_prepare_nodes = 25001
sim_mplanintparam_clear_nodes = 25002

# User interface elements
sim_gui_menubar = 0x0001
sim_gui_popups = 0x0002
sim_gui_toolbar1 = 0x0004
sim_gui_toolbar2 = 0x0008
sim_gui_hierarchy = 0x0010
sim_gui_infobar = 0x0020
sim_gui_statusbar = 0x0040
sim_gui_scripteditor = 0x0080
sim_gui_scriptsimulationparameters = 0x0100
sim_gui_dialogs = 0x0200
sim_gui_browser = 0x0400
sim_gui_all = 0xffff


# Joint modes
sim_jointmode_passive = 0
sim_jointmode_motion = 1
sim_jointmode_ik = 2
sim_jointmode_ikdependent = 3
sim_jointmode_dependent = 4
sim_jointmode_force = 5


# Navigation and selection modes with the mouse. Lower byte values are mutually exclusive upper byte bits can be combined
sim_navigation_passive = 0x0000
sim_navigation_camerashift = 0x0001
sim_navigation_camerarotate = 0x0002
sim_navigation_camerazoom = 0x0003
sim_navigation_cameratilt = 0x0004
sim_navigation_cameraangle = 0x0005
sim_navigation_camerafly = 0x0006
sim_navigation_objectshift = 0x0007
sim_navigation_objectrotate = 0x0008
sim_navigation_reserved2 = 0x0009
sim_navigation_reserved3 = 0x000A
sim_navigation_jointpathtest = 0x000B
sim_navigation_ikmanip = 0x000C
sim_navigation_objectmultipleselection = 0x000D
# Bit-combine following values and add them to one of above's values for a valid navigation mode
sim_navigation_reserved4 = 0x0100
sim_navigation_clickselection = 0x0200
sim_navigation_ctrlselection = 0x0400
sim_navigation_shiftselection = 0x0800
sim_navigation_camerazoomwheel = 0x1000
sim_navigation_camerarotaterightbutton = 0x2000


# Remote API constants
SIMX_VERSION = 0
# Remote API message header structure
SIMX_HEADER_SIZE = 18
# 1 simxUShort. Generated by the client or server. The CRC for the message
simx_headeroffset_crc = 0
# 1 byte. Generated by the client or server. The version of the remote API software
simx_headeroffset_version = 2
# 1 simxInt. Generated by the client (and used in a reply by the server)
simx_headeroffset_message_id = 3
# 1 simxInt. Client time stamp generated by the client (and sent back by the server)
simx_headeroffset_client_time = 7
# 1 simxInt. Generated by the server when a reply is generated. The server timestamp
simx_headeroffset_server_time = 11
# 1 simxUShort. Generated by the server. A unique ID identifying the scene currently displayed
simx_headeroffset_scene_id = 15
# 1 byte. Generated by the server. Bit coded 0 set --> simulation not stopped 1 set --> simulation paused 2 set --> real-time switch on 3-5 edit mode type (0=no edit mode 1=triangle 2=vertex 3=edge 4=path 5=UI)
simx_headeroffset_server_state = 17

# Remote API command header
SIMX_SUBHEADER_SIZE = 26
# 1 simxInt. Generated by the client or server. The buffer size of the command.
simx_cmdheaderoffset_mem_size = 0
# 1 simxInt. Generated by the client or server. The full buffer size of the command (applies to split chunks).
simx_cmdheaderoffset_full_mem_size = 4
# 1 simxUShort. Generated by the client or server. The amount of data that is part of the command identification.
simx_cmdheaderoffset_pdata_offset0 = 8
# 1 simxInt. Generated by the client or server. The amount of shift of the pure data buffer (applies to split chunks).
simx_cmdheaderoffset_pdata_offset1 = 10
# 1 simxInt. Generated by the client (and used in a reply by the server). The command combined with the operation mode of the command.
simx_cmdheaderoffset_cmd = 14
# 1 simxUShort. Generated by the client or server. The amount of delay in ms of a continuous command or the max. pure data size to send at once (applies to split commands).
simx_cmdheaderoffset_delay_or_split = 18
# 1 simxInt. Generated by the server. The simulation time (in ms) when the command was executed (or 0 if simulation is not running)
simx_cmdheaderoffset_sim_time = 20
# 1 byte. Generated by the server. (1 bit 0 is set --> error in function execution on server side). The client writes bit 1 if command cannot be overwritten
simx_cmdheaderoffset_status = 24
simx_cmdheaderoffset_reserved = 25    # 1 byte. Not yet used


# Regular operation modes
# sends command as one chunk. Reply will also come as one chunk. Doesn't wait for the reply.
simx_opmode_oneshot = 0x000000
# sends command as one chunk. Reply will also come as one chunk. Waits for the reply (_REPLY_WAIT_TIMEOUT_IN_MS is the timeout).
simx_opmode_blocking = 0x010000
# sends command as one chunk. Reply will also come as one chunk. Waits for the reply (_REPLY_WAIT_TIMEOUT_IN_MS is the timeout).
simx_opmode_oneshot_wait = 0x010000
simx_opmode_continuous = 0x020000
# sends command as one chunk. Command will be stored on the server and always executed
simx_opmode_streaming = 0x020000
# (every x ms (as far as possible) where x can be 0-65535. just add x to opmode_continuous).
# A reply will be sent continuously each time as one chunk. Doesn't wait for the reply.

# Operation modes for heavy data
# sends command as several chunks (max chunk size is x bytes where x can be _MIN_SPLIT_AMOUNT_IN_BYTES-65535. Just add x to opmode_oneshot_split). Reply will also come as several chunks. Doesn't wait for the reply.
simx_opmode_oneshot_split = 0x030000
simx_opmode_continuous_split = 0x040000
# sends command as several chunks (max chunk size is x bytes where x can be _MIN_SPLIT_AMOUNT_IN_BYTES-65535. Just add x to opmode_continuous_split). Command will be stored on the server and always executed. A reply will be sent continuously each time as several chunks. Doesn't wait for the reply.
simx_opmode_streaming_split = 0x040000

# Special operation modes
# removes and cancels all commands stored on the client or server side (also continuous commands)
simx_opmode_discontinue = 0x050000
# doesn't send anything but checks if a reply for the given command is available in the input buffer (i.e. previously received from the server)
simx_opmode_buffer = 0x060000
# doesn't send anything and doesn't return any specific value. It just erases a similar command reply in the inbox (to free some memory)
simx_opmode_remove = 0x070000


# Command return codes
simx_return_ok = 0x000000
# input buffer doesn't contain the specified command
simx_return_novalue_flag = 0x000001
# command reply not received in time for opmode_oneshot_wait operation mode
simx_return_timeout_flag = 0x000002
# command doesn't support the specified operation mode
simx_return_illegal_opmode_flag = 0x000004
# command caused an error on the server side
simx_return_remote_error_flag = 0x000008
# previous similar command not yet fully processed (applies to opmode_oneshot_split operation modes)
simx_return_split_progress_flag = 0x000010
# command caused an error on the client side
simx_return_local_error_flag = 0x000020
simx_return_initialize_error_flag = 0x000040        # simxStart was not yet called

# Following for backward compatibility (same as above)
simx_error_noerror = 0x000000
# input buffer doesn't contain the specified command
simx_error_novalue_flag = 0x000001
# command reply not received in time for opmode_oneshot_wait operation mode
simx_error_timeout_flag = 0x000002
# command doesn't support the specified operation mode
simx_error_illegal_opmode_flag = 0x000004
# command caused an error on the server side
simx_error_remote_error_flag = 0x000008
# previous similar command not yet fully processed (applies to opmode_oneshot_split operation modes)
simx_error_split_progress_flag = 0x000010
# command caused an error on the client side
simx_error_local_error_flag = 0x000020
simx_error_initialize_error_flag = 0x000040        # simxStart was not yet called



from stereo3d.stereocapture.pyloncapture import PylonCapture
import numpy as np
import cv2


class StereoCapturePylon():
    def __init__(self, camL, camR):
        """
        Initialisation function for StereoCapturePylon class.
        :param camL: left camera
        :param camR: right camera
        :type camL: PylonCapture
        :type camR: PylonCapture
        """
        self.camL = camL
        self.camR = camR

    def connect(self):
        """
        Connect to cameras.
        :returns: success of connection
        :rtype: bool
        """
        resL = self.camL.connect()
        resR = self.camR.connect()
        res = resL and resR
        return res

    def grab(self):
        """
        Grab images from each camera in stereo pair
        :returns: success of capture, image left, image right
        :rtype: bool, numpy.array, numpy.array
        """
        resL, image_left = self.camL.grab()
        resR, image_right = self.camR.grab()
        res = resL and resR
        return res, image_left, image_right

    def close(self):
        """
        Close connection to cameras
        """
        self.camL.close()
        self.camR.close()


if __name__ == "__main__":
    left_camera_serial = "23517286"
    right_camera_serial = "23517294"
    camL = PylonCapture(left_camera_serial)
    camR = PylonCapture(right_camera_serial)
    stcam = StereoCapturePylon(camL, camR)
    stcam.connect()
    while True:
        res, imageL, imageR = stcam.grab()
        if res:
            stereo_image = np.concatenate((imageL, imageR), axis=1)
            stereo_image_resized = cv2.resize(stereo_image, (1280, 480))
            cv2.imshow('Stereo Image', stereo_image_resized)
            k = cv2.waitKey(1)
            if k == ord('q'):
                break
    stcam.close()



"""Stereo3D module"""
import os
import sys
import time
import numpy as np
import cv2
from pymsgbox import prompt, alert
from pyntcloud import PyntCloud
from i3drsgm import I3DRSGM
import pandas as pd
import traceback


class Stereo3D():
    def __init__(self, stereo_camera, stereo_calibration,
                 stereo_matcher=None, show_window=True):
        """
        Initialisation function for PickPlace3D. Defines the devices used.
        If you would like to use the methods without connecting
        a camera then initaialise Stereo3D as 's3D = Stereo3D()'.
        :param left_cal_file:
            filepath to left image calibration file (e.g. left.yaml)
        :param right_cal_file:
            filepath to right image calibration file (e.g. right.yaml)
        :param stereo_camera:
            stereo camera used for generating 3D and 2D detection
        :type left_cal_file: string
        :type right_cal_file: string
        :type stereo_camera: StereoCapture.StereoCapture
        """
        self.change_camera(stereo_camera)
        self.stereo_calibration = stereo_calibration
        self.Q = self.stereo_calibration.stereo_cal["q"]

        self.image_left = None
        self.image_right = None
        self.rect_image_left = None
        self.rect_image_right = None
        self.disparity = None
        self.depth = None

        self.EXIT_CODE_QUIT = -1
        self.EXIT_CODE_GRAB_3D_SUCCESS = 1
        self.EXIT_CODE_FAILED_TO_GRAB_3D = -2

        self.save_index = 0

        self.cv_window_name_Controls = "[Stereo3D] Controls"
        self.cv_window_name_Images = "[Stereo3D] Images"

        self.show_window = show_window
        if self.show_window:
            cv2.namedWindow(self.cv_window_name_Controls, cv2.WINDOW_NORMAL)
            # cv2.namedWindow(self.cv_window_name_Images, cv2.WINDOW_NORMAL)

            cv2.setMouseCallback(
                self.cv_window_name_Images, self.on_window_mouse)

            cv2.resizeWindow(self.cv_window_name_Controls, 400, 0)

        self.matcher_name = ""
        self.change_matcher(stereo_matcher)

        # Once init has been called matcher can be changed e.g.
        # s3D = Stereo3D()
        # s3D.ply_header = "ply..."
        self.ply_header = (
            "ply\n"
            "format ascii 1.0\n"
            "element vertex %(vert_num)d\n"
            "property float x\n"
            "property float y\n"
            "property float z\n"
            "property uchar red\n"
            "property uchar green\n"
            "property uchar blue\n"
            "end_header\n")

        self.frame_count = 0

    def change_camera(self, stereo_camera):
        self.stereo_camera = stereo_camera

    def change_matcher(self, stereo_matcher):
        default_min_disp = 1000
        default_num_disparities = 20
        default_block_size = 12
        default_uniqueness_ratio = 15
        default_texture_threshold = 15
        default_speckle_size = 0
        default_speckle_range = 500

        if (stereo_matcher is None):
            self.matcher = cv2.StereoBM_create()
            calc_block = (2 * default_block_size + 5)
            self.matcher.setBlockSize(calc_block)
            self.matcher.setMinDisparity(int(default_min_disp - 1000))
            self.matcher.setNumDisparities(16*(default_num_disparities+1))
            self.matcher.setUniquenessRatio(default_uniqueness_ratio)
            self.matcher.setTextureThreshold(default_texture_threshold)
            self.matcher.setSpeckleWindowSize(default_speckle_size)
            self.matcher.setSpeckleRange(default_speckle_range)
        else:
            if stereo_matcher == "BM":
                self.matcher_name = stereo_matcher
                self.matcher = cv2.StereoBM_create()
                calc_block = (2 * default_block_size + 5)
                self.matcher.setBlockSize(calc_block)
                self.matcher.setMinDisparity(int(default_min_disp - 1000))
                self.matcher.setNumDisparities(16*(default_num_disparities+1))
                self.matcher.setUniquenessRatio(default_uniqueness_ratio)
                self.matcher.setTextureThreshold(default_texture_threshold)
                self.matcher.setSpeckleWindowSize(default_speckle_size)
                self.matcher.setSpeckleRange(default_speckle_range)
            elif stereo_matcher == "SGBM":
                self.matcher_name = stereo_matcher
                self.matcher = cv2.StereoSGBM_create()
                calc_block = (2 * default_block_size + 5)
                self.matcher.setBlockSize(calc_block)
                self.matcher.setMinDisparity(int(default_min_disp - 1000))
                self.matcher.setNumDisparities(16*(default_num_disparities+1))
                self.matcher.setUniquenessRatio(default_uniqueness_ratio)
                # self.matcher.setTextureThreshold(default_texture_threshold)
                self.matcher.setSpeckleWindowSize(default_speckle_size)
                self.matcher.setSpeckleRange(default_speckle_range)
            elif stereo_matcher == "I3DRSGM":
                self.matcher_name = stereo_matcher
                self.matcher = I3DRSGM()
                self.matcher.setWindowSize(11)
                self.matcher.setMinDisparity(0)
                self.matcher.setDisparityRange(16*120)
                self.matcher.setPyamidLevel(6)
            else:
                self.matcher = stereo_matcher

        if self.show_window:
            cv2.destroyWindow(self.cv_window_name_Controls)
            cv2.namedWindow(self.cv_window_name_Controls, cv2.WINDOW_NORMAL)

            cv2.createTrackbar(
                "Min disp", self.cv_window_name_Controls,
                default_min_disp, 2000, self.on_min_disparity_trackbar)
            cv2.createTrackbar(
                "Disp", self.cv_window_name_Controls,
                default_num_disparities, 30, self.on_num_disparities_trackbar)
            cv2.createTrackbar(
                "Blck sze", self.cv_window_name_Controls,
                default_block_size, 100, self.on_block_size_trackbar)

            cv2.createTrackbar(
                "Uniq", self.cv_window_name_Controls,
                default_uniqueness_ratio, 100,
                self.on_uniqueness_ratio_trackbar)
            if stereo_matcher == "BM":
                cv2.createTrackbar(
                    "Texture", self.cv_window_name_Controls,
                    default_texture_threshold, 100,
                    self.on_texture_threshold_trackbar)

            cv2.createTrackbar(
                "Sp size", self.cv_window_name_Controls,
                default_speckle_size, 500, self.on_speckle_size_trackbar)
            cv2.createTrackbar(
                "Sp range", self.cv_window_name_Controls,
                default_speckle_range, 1000, self.on_speckle_range_trackbar)

    def connect(self):
        """
        Connect to devices needed for the pick and place process.
        :returns: success
        :rtype: bool
        """
        res = self.stereo_camera.connect()
        return res

    def gen3D(self, left_image, right_image):
        if (self.matcher_name == "I3DRSGM"):
            valid, disparity = self.matcher.forwardMatch(left_image, right_image)
            disparity = -disparity.astype(np.float32)
            disparity[disparity==99999]=0.0
            disparity[disparity<=0]=0.0
            disparity = np.nan_to_num(disparity, nan=0.0,posinf=0.0,neginf=0.0)
        else:
            disparity = self.matcher.compute(left_image, right_image)
            disparity = disparity.astype(np.float32) / 16.0
        return disparity

    def genDepth(self, disparity):
        depth = cv2.reprojectImageTo3D(disparity, self.Q)
        return depth

    def write_ply(self, filename, disp, depth, image):
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = np.float32(image)
        image = image/255

        mask = disp > disp.min()
        points = depth[mask]
        image = image[mask]

        # TODO replace this with better filter in-case
        # 3D is every large an this becomes invalid
        points[points >= 100] = 0
        points[points <= -100] = 0

        points = points.reshape(-1, 3)
        image = image.reshape(-1, 3)

        points = np.hstack((points, image))

        cloud = PyntCloud(pd.DataFrame(
            # same arguments that you are passing to visualize_pcl
            data=points,
            columns=["x", "y", "z", "red", "green", "blue"]))

        cloud.to_file(filename)

    def scale_disparity(self, disparity):
        minV, maxV, _, _ = cv2.minMaxLoc(disparity)
        if (maxV - minV != 0):
            scaled_disp = cv2.convertScaleAbs(
                disparity, alpha=255.0/(maxV - minV),
                beta=-minV * 255.0/(maxV - minV))
            return scaled_disp
        else:
            return np.zeros(disparity.shape, np.uint8)

    def on_min_disparity_trackbar(self, val):
        min_disp = int(val - 1000)
        self.matcher.setMinDisparity(min_disp)

    def on_block_size_trackbar(self, val):
        if (self.matcher_name == "I3DRSGM"):
            self.matcher.setWindowSize(2 * val + 5)
        else:
            self.matcher.setBlockSize(2 * val + 5)

    def on_num_disparities_trackbar(self, val):
        if (self.matcher_name == "I3DRSGM"):
            self.matcher.setDisparityRange(16*((val*10)+1))
        else:
            self.matcher.setNumDisparities(16*(val+1))

    def on_texture_threshold_trackbar(self, val):
        if (self.matcher_name != "I3DRSGM"):
            self.matcher.setTextureThreshold(val)

    def on_uniqueness_ratio_trackbar(self, val):
        if (self.matcher_name != "I3DRSGM"):
            self.matcher.setUniquenessRatio(val)

    def on_speckle_size_trackbar(self, val):
        if (self.matcher_name != "I3DRSGM"):
            self.matcher.setSpeckleWindowSize(val)

    def on_speckle_range_trackbar(self, val):
        if (self.matcher_name != "I3DRSGM"):
            self.matcher.setSpeckleRange(val)

    def on_window_mouse(self, event, x, y, flags, param):
        if event == cv2.EVENT_LBUTTONDOWN:
            print("click")

    def grab3D(self, isRectified=False):
        res, image_left, image_right = self.stereo_camera.grab()
        if (res):
            self.image_left = image_left
            self.image_right = image_right
            if (isRectified):
                self.rect_image_left = self.image_left
                self.rect_image_right = self.image_right
            else:
                self.rect_image_left, self.rect_image_right = \
                    self.stereo_calibration.rectify_pair(
                        image_left, image_right)

            disp = self.gen3D(self.rect_image_left, self.rect_image_right)
            self.disparity = disp

            return True, disp
        else:
            print("Failed to grab stereo image pair")
            return False, None

    def save_point_cloud(self, disparity, image,
                         defaultSaveFolder="", points_file_string="output.ply",
                         confirm_folder=True):
        # prompt user for save location
        resp = defaultSaveFolder
        if (confirm_folder):
            try:
                resp = prompt(
                    text='Saving 3D Point Cloud to path: ',
                    title='Save 3D Point Cloud', default=defaultSaveFolder)
            except AssertionError:
                _, _, tb = sys.exc_info()
                traceback.print_tb(tb)  # Fixed format
                tb_info = traceback.extract_tb(tb)
                filename, line, func, text = tb_info[-1]

                err_msg = 'An error occurred on line {} in statement {}'
                err_msg = err_msg.format(line, text)
                print(err_msg)
                err_msg = 'Will use default save folder: {}'
                err_msg = err_msg.format(defaultSaveFolder)
                print(err_msg)
        if (resp is not None and disparity is not None and image is not None):
            # define name of output point cloud ply file
            ply_filename = os.path.join(resp, points_file_string)

            # generate depth from disparity
            print("Generating depth from disparity...")
            depth = self.genDepth(disparity)
            print("Saving point cloud...")
            # write 3D data to ply with color from image on points
            self.write_ply(ply_filename, disparity, depth, image)
            print("Point cloud save complete.")
            if (confirm_folder):
                alert('3D point cloud saved.', 'Save 3D Point Cloud')
        else:
            print("invalid prompt response or disparity/image is empty")

    def save_all_current(self, saveFolder):
        confirm_folder = False
        left_file_string = str(self.save_index)+"_l.png"
        right_file_string = str(self.save_index)+"_r.png"
        self.stereo_camera.save_images(
            self.image_left, self.image_right,
            saveFolder, left_file_string, right_file_string,
            confirm_folder)

        left_file_string = "rect_"+str(self.save_index)+"_l.png"
        right_file_string = "rect_"+str(self.save_index)+"_r.png"
        self.stereo_camera.save_images(
            self.rect_image_left, self.rect_image_right,
            saveFolder, left_file_string, right_file_string,
            confirm_folder)

        points_file_string = "points_"+str(self.save_index)+".ply"
        self.save_point_cloud(
            self.disparity, self.rect_image_left,
            saveFolder, points_file_string, confirm_folder)

    def run_frame_no_gui(self, saveFolder, isRectified=False):
        # grab 3D disparity from stereo camera
        res, _ = self.grab3D(isRectified)
        if res:
            self.save_all_current(saveFolder)
        return res

    def run_frame(self, defaultSaveFolder="", isRectified=False,
                  confirm_folder=True, colormap=cv2.COLORMAP_JET):
        # grab 3D disparity from stereo camera
        res, disp = self.grab3D(isRectified)
        if res:
            self.frame_count += 1
            # prepare images for displaying
            display_image = np.zeros((640, 480), np.uint8)

            rect_image_left_resized = \
                self.stereo_camera.image_resize(
                    self.rect_image_left, height=640)
            rect_image_right_resized = \
                self.stereo_camera.image_resize(
                    self.rect_image_right, height=640)

            disp_resized = self.scale_disparity(
                self.stereo_camera.image_resize(disp, height=640))
            disp_black_mask = disp_resized <= 0
            # apply color map to disparity
            disp_colormap = cv2.applyColorMap(disp_resized, colormap)
            disp_colormap[disp_black_mask != 0] = [0, 0, 0]

            left_right_dual_gray = np.concatenate(
                (rect_image_left_resized, rect_image_right_resized), axis=1)
            left_right_dual = cv2.cvtColor(
                left_right_dual_gray, cv2.COLOR_GRAY2RGB)

            (lr_dual_h, lr_dual_w, _) = left_right_dual.shape
            (d_h, d_w, _) = disp_colormap.shape

            spacer_width_raw = (lr_dual_w - d_w)
            if (spacer_width_raw % 2) == 0:
                # even
                spacer_width_1 = int((spacer_width_raw / 2))
                spacer_width_2 = int((spacer_width_raw / 2))
            else:
                # odd
                spacer_width_1 = int((spacer_width_raw / 2))
                spacer_width_2 = int((spacer_width_raw / 2) + 1)

            disp_spacer_1 = np.zeros((640, spacer_width_1, 3), np.uint8)
            disp_spacer_2 = np.zeros((640, spacer_width_2, 3), np.uint8)
            disp_spaced = np.concatenate(
                (disp_spacer_1, disp_colormap, disp_spacer_2), axis=1)

            display_image = np.concatenate(
                (disp_spaced, left_right_dual), axis=0)
            display_image_resize = self.stereo_camera.image_resize(
                display_image, height=640)

            font = cv2.FONT_HERSHEY_DUPLEX
            bottomLeftCornerOfText = (10, 20)
            fontScale = 0.4
            fontColor = (255, 255, 255)
            lineType = 1

            cv2.putText(
                display_image_resize, 'Frame: {}'.format(self.frame_count),
                bottomLeftCornerOfText,
                font,
                fontScale,
                fontColor,
                lineType)

            if self.show_window:
                # display disparity with stereo images
                cv2.imshow(self.cv_window_name_Images, display_image_resize)

        k = cv2.waitKey(1)
        if k == ord('q'):  # exit if 'q' key pressed
            return self.EXIT_CODE_QUIT
        elif k == ord('s'):  # save stereo image pair
            left_file_string = str(self.save_index)+"_l.png"
            right_file_string = str(self.save_index)+"_r.png"
            self.stereo_camera.save_images(
                self.image_left, self.image_right,
                defaultSaveFolder, left_file_string, right_file_string,
                confirm_folder)
            self.save_index += 1
        elif k == ord('r'):  # save rectified stereo image pair
            left_file_string = "rect_"+str(self.save_index)+"_l.png"
            right_file_string = "rect_"+str(self.save_index)+"_r.png"
            self.stereo_camera.save_images(
                self.rect_image_left, self.rect_image_right,
                defaultSaveFolder, left_file_string, right_file_string,
                confirm_folder)
            self.save_index += 1
        elif k == ord('p'):  # save 3D data as point cloud
            points_file_string = "points_"+str(self.save_index)+".ply"
            self.save_point_cloud(
                self.disparity, self.rect_image_left,
                defaultSaveFolder, points_file_string,
                confirm_folder)
            self.save_index += 1
        elif k == ord('1'):  # change tp OpenCV BM
            self.change_matcher("BM")
        elif k == ord('2'):  # change to OpenCV SGBM
            self.change_matcher("SGBM")
        if self.show_window:
            window_image_prop = cv2.getWindowProperty(
                self.cv_window_name_Images, cv2.WND_PROP_VISIBLE)
            window_control_prop = cv2.getWindowProperty(
                self.cv_window_name_Images, cv2.WND_PROP_VISIBLE)
            if window_image_prop < 1:
                return self.EXIT_CODE_QUIT
            if window_control_prop < 1:
                return self.EXIT_CODE_QUIT

        if res:
            return self.EXIT_CODE_GRAB_3D_SUCCESS
        else:
            return self.EXIT_CODE_FAILED_TO_GRAB_3D

    def run(self, defaultSaveFolder="", isRectified=False,
            frame_delay=0, confirm_folder=True, colormap=cv2.COLORMAP_JET):
        # connect to stereo camera
        connected = False
        while(not connected):
            connected = self.connect()
            time.sleep(1)
        if self.show_window:
            cv2.namedWindow(self.cv_window_name_Images, cv2.WINDOW_NORMAL)
        while(True):
            exit_code = self.run_frame(
                defaultSaveFolder, isRectified, confirm_folder, colormap)
            if (exit_code == self.EXIT_CODE_QUIT):
                break
            if (exit_code == self.EXIT_CODE_FAILED_TO_GRAB_3D):
                time.sleep(1)
            time.sleep(frame_delay)

        self.stereo_camera.close()



# TODO TIS module is linux only so class not included in the package untill windows support is added

import cv2
import numpy as np
import os
import TIS
import time
from collections import namedtuple

# This sample shows, how to get an image in a callback and use trigger or software trigger
# needed packages:
# python-opencv
# python-gst-1.0
# tiscamera

class CustomData:
        ''' Example class for user data passed to the on new image callback function
        '''

        def __init__(self, newImageReceived, image):
                self.newImageReceived = newImageReceived
                self.image = image
                self.busy = False

class ICCapture():
    def __init__(self,serial,width,height,fps,color=False):
        self.serial = serial
        self.width = width
        self.height = height
        self.fps = fps
        self.color = color

    def connect(self):
        self.CD = CustomData(False, None)
        # Open camera, set video format, framerate and determine, whether the sink is color or bw
        # Parameters: Serialnumber, width, height, framerate (numerator only) , color
        # If color is False, then monochrome / bw format is in memory. If color is True, then RGB32
        # colorformat is in memory
        self.Tis = TIS.TIS(self.serial, self.width, self.height, self.fps, False)

        self.Tis.Set_Image_Callback(self.on_new_image, self.CD)

        # Tis.Set_Property("Trigger Mode", "Off") # Use this line for GigE cameras
        self.Tis.Set_Property("Trigger Mode", False)
        self.CD.busy = True # Avoid, that we handle image, while we are in the pipeline start phase
        # Start the pipeline
        print("starting pipeline...")
        self.Tis.Start_pipeline()
        print("pipeline setup")

        #self.Tis.List_Properties()

        # Tis.Set_Property("Trigger Mode", "On") # Use this line for GigE cameras
        self.Tis.Set_Property("Trigger Mode", False)
        cv2.waitKey(1000)

        self.CD.busy = False  # Now the callback function does something on a trigger


        # Query the gain auto and current value :
        print("Gain Auto : %s " % self.Tis.Get_Property("Gain Auto").value)
        print("Gain : %d" % self.Tis.Get_Property("Gain").value)

        # Check, whether gain auto is enabled. If so, disable it.

        if self.Tis.Get_Property("Gain Auto").value :
                self.Tis.Set_Property("Gain Auto",False)
                print("Gain Auto now : %s " % self.Tis.Get_Property("Gain Auto").value)

        self.Tis.Set_Property("Gain",0)

        # Now do the same with exposure. Disable automatic if it was enabled
        # then set an exposure time.
        if self.Tis.Get_Property("Exposure Auto").value :
                self.Tis.Set_Property("Exposure Auto", False)
                print("Exposure Auto now : %s " % self.Tis.Get_Property("Exposure Auto").value)

        self.Tis.Set_Property("Exposure Time (us)", 15000)

        print("Exposure Auto : %s " % self.Tis.Get_Property("Exposure Auto").value)
        print("Exposure : %d" % self.Tis.Get_Property("Exposure Time (us)").value)

    def grab(self):
        #time.sleep(1)
        self.Tis.Set_Property("Software Trigger",1) # Send a software trigger

        # Wait for a new image. Use 10 tries.
        tries = 10
        while self.CD.newImageReceived is False and tries > 0:
                time.sleep(0.1)
                tries -= 1

        # Check, whether there is a new image and handle it.
        if self.CD.newImageReceived is True:
                self.CD.newImageReceived = False
                image = self.CD.image
                return True,image
                #cv2.imshow('Window', CD.image)
        else:
                print("No image received")
                return False,None

    def on_new_image(self,tis, userdata):
        '''
        Callback function, which will be called by the TIS class
        :param tis: the camera TIS class, that calls this callback
        :param userdata: This is a class with user data, filled by this call.
        :return:
        '''
        # Avoid being called, while the callback is busy
        if userdata.busy is True:
                return

        userdata.busy = True
        userdata.newImageReceived = True
        userdata.image = tis.Get_image()
        userdata.busy = False
        
    def close(self):
        self.Tis.Stop_pipeline()

if __name__ == '__main__':
    camera_serial = "27810457"
    cam = ICCapture(camera_serial, 2448, 2048, 35)
    cam.connect()
    while(True):
        res, image = cam.grab()
        if (res):
            image_resized = cv2.resize(image,(640,480))
            cv2.imshow('Image', image_resized)
            k = cv2.waitKey(1)
            if k == ord('q'):
                break
    cam.close()


import cv2


class CVCapture():
    def __init__(self, camera_index):
        """
        Initialisation function for CVCapture class.
        :param camera_index: id of the video capturing device to open
        :type camera_index: int
        """
        self.camera = None
        self.camera_index = camera_index

    def connect(self):
        """
        Connect to camera using opencv VideoCaputre
        :returns: success of connection
        :rtype: bool
        """
        self.camera = cv2.VideoCapture(self.camera_index)
        return True

    def grab(self):
        """
        Grab image from camera
        :returns: success of capture, image
        :rtype: bool, numpy.array
        """
        ret, image = self.camera.read()
        return ret, image

    def close(self):
        """
        Close connection to camera
        """
        self.camera.release()


if __name__ == '__main__':
    CAMERA_INDEX = 0
    cam = CVCapture(CAMERA_INDEX)
    cam.connect()
    while True:
        res, image = cam.grab()
        if res:
            image_resized = cv2.resize(image, (640, 480))
            cv2.imshow('Image', image_resized)
            k = cv2.waitKey(1)
            if k == ord('q'):
                break
    cam.close()



__author__ = "Ben Knight @ Industrial 3D Robotics"
__maintainer__ = "Ben Knight"
__email__ = "bknight@i3drobotics.com"
__copyright__ = "Copyright 2020, Industrial 3D Robotics"
__license__ = "MIT"
__docformat__ = 'reStructuredText'

from pymsgbox import prompt, alert
import numpy as np
import cv2
from stereo3d.stereocapture.pyloncapture import PylonCapture
from stereo3d.stereocapture.cvcapture import CVCapture
from stereo3d.stereocapture.cvimagecapture import CVImageCapture
from stereo3d.stereocapture.stereocapturepylon import StereoCapturePylon
from stereo3d.stereocapture.stereocapturecvsplit import StereoCaptureCVSplit
from stereo3d.stereocapture.stereocapturecvdual import StereoCaptureCVDual


class StereoCapture():
    """
    Python tool for capturing stereo pairs from difference sources
    See README for details on usages:
    https://github.com/i3drobotics/StereoCapture/blob/master/pyStereoCapture/StereoCapture/README.md
    """
    def __init__(self, stcam, param=None):
        """
        Initialisation function for StereoCapture class.
        :param cam:
            camera to use as stereo device.
            This should one of three types:
                - StereoCapturePylon
                - StereoCaptureCVSplit
                - StereoCaptureCVDual
        Some prebuild stereo cameras can be used by using:
                "Deimos"/"Phobos"/"Image"/"Pylon"
        :type cam:
            - StereoCapturePylon
            - StereoCaptureCVSplit
            - StereoCaptureCVDual
            - String("Deimos"/"Phobos"/"Image"/"Pylon")
        """
        self.stcam = None
        if (stcam == "Deimos" or stcam == "deimos"):
            if param is not None:
                print("Setting up as deimos stereo camera...")
                cam = CVCapture(param)
                stcam = StereoCaptureCVDual(cam)
            else:
                err_msg = "param MUST be defined when "
                err_msg += "using pre-made stereo capture object. "
                err_msg += "(int)usb_camera_index"
                print(err_msg)
                return
        elif (stcam == "Phobos" or stcam == "phobos"
              or stcam == "Pylon" or stcam == "pylon"):
            print("Setting up as phobos stereo camera...")
            if (param is not None):
                left_camera_serial = param[0]
                right_camera_serial = param[1]
                camL = PylonCapture(left_camera_serial, trigger_mode=True)
                camR = PylonCapture(right_camera_serial, trigger_mode=True)
                stcam = StereoCapturePylon(camL, camR)
            else:
                err_msg = "param MUST be defined when using pre-made "
                err_msg += "stereo capture object. "
                err_msg += "(Array)[Left_serial,Right_serial]"
                print(err_msg)
                return
        elif (stcam == "Image" or stcam == "image"):
            print("Setting up as image stereo camera...")
            if (param is not None):
                # TODO check files exist
                camL = CVImageCapture(param[0])
                camR = CVImageCapture(param[1])
                stcam = StereoCaptureCVSplit(camL, camR)
            else:
                err_msg = "param MUST be defined when using pre-made "
                err_msg += "stereo capture object. "
                err_msg += "(Array)[Left_serial,Right_serial]"
                print(err_msg)
        self.stcam = stcam
        self.flip_h = False
        self.flip_v = False

    def connect(self):
        """
        Connect to stereo camera.
        :returns: success of connection
        :rtype: bool
        """
        if (self.stcam is not None):
            res = self.stcam.connect()
        else:
            print("Stereo camera not defined")
            res = False
        return res

    def grab(self):
        """
        Grab images from stereo camera
        :returns: success of capture, image left, image right
        :rtype: bool, numpy.array, numpy.array
        """
        if (self.stcam is not None):
            res, imageL, imageR = self.stcam.grab()
            if (self.flip_h):
                imageL = self.flip_image_h(imageL)
                imageR = self.flip_image_h(imageR)
            if (self.flip_v):
                imageL = self.flip_image_v(imageL)
                imageR = self.flip_image_v(imageR)
        else:
            print("Stereo camera is not defined")
            res = False
        return res, imageL, imageR

    def flip_image_h(self, image):
        flipImage = cv2.flip(image, 1)
        return flipImage

    def flip_image_v(self, image):
        flipImage = cv2.flip(image, 0)
        return flipImage

    def save_images(self, image_left, image_right,
                    defaultSaveFolder="", left_file_string="left.png",
                    right_file_string="right.png", confirm_folder=True):
        """
        Save stereo images to files
        :param image_left: left camera image matrix
        :param image_right: right camera image matrix
        :param defaultSaveFolder:
            default folder to save the images to
            (will still ask for confirmation)
        :param left_file_string: left image filename
        :param right_file_string: right image filename
        :type image_left: numpy
        :type image_right: numpy
        :type defaultSaveFolder: string
        :type left_file_string: string
        :type right_file_string: string
        """
        # prompt user for save location
        if (confirm_folder):
            resp = prompt(
                text='Saving image pair to path: ',
                title='Save Image Pair', default=defaultSaveFolder)
        else:
            resp = defaultSaveFolder
        if (None not in [resp, image_left, image_right]):
            # define name of output images
            left_image_filename = resp + left_file_string
            right_image_filename = resp + right_file_string

            print("Saving stereo image pair...")
            cv2.imwrite(left_image_filename, image_left)
            cv2.imwrite(right_image_filename, image_right)
            print("Stereo image pair saved")
            if (confirm_folder):
                alert('Stereo image pair saved.', 'Save Image Pair')
        else:
            print("Invalid prompt response or images are empty")

    def image_resize(self, image, width=None, height=None,
                     inter=cv2.INTER_AREA):
        """
        Resize image based on height or width while maintaning aspect ratio
        :param image: image matrix
        :param width:
            desired width of output image
            (can only use width or height not both)
        :param height:
            desired height of output image
            (can only use width or height not both)
        :param inter: opencv resize method (default: cv2.INTER_AREA)
        :type image: numpy
        :type width: int
        :type height: int
        :type inter: int
        """
        # initialize the dimensions of the image to be resized and
        # grab the image size
        dim = None
        (h, w) = image.shape[:2]

        # if both the width and height are None, then return the
        # original image
        if width is None and height is None:
            return image

        # check to see if the width is None
        if width is None:
            # calculate the ratio of the height and construct the
            # dimensions
            r = height / float(h)
            dim = (int(w * r), height)

        # otherwise, the height is None
        else:
            # calculate the ratio of the width and construct the
            # dimensions
            r = width / float(w)
            dim = (width, int(h * r))

        # resize the image
        resized = cv2.resize(image, dim, interpolation=inter)

        # return the resized image
        return resized

    def runGui(self, defaultSaveFolder="", confirm_folder=True):
        """
        Display GUI for viewing stereo camera feed
        """
        image_right = None
        image_left = None
        if (self.stcam is not None):
            self.connect()
            save_index = 0
            while(True):
                res, in_image_left, in_image_right = self.stcam.grab()
                if (res):
                    image_right = in_image_right
                    image_left = in_image_left
                    stereo_image = np.concatenate(
                        (image_left, image_right), axis=1)
                    stereo_image_resized = self.image_resize(
                        stereo_image, 1280)
                    cv2.imshow('Stereo Image', stereo_image_resized)
                k = cv2.waitKey(1)
                if k == ord('q'):
                    break
                elif k == ord('s'):  # save stereo image pair
                    left_file_string = str(save_index)+"_l.png"
                    right_file_string = str(save_index)+"_r.png"
                    self.save_images(
                        image_left, image_right, defaultSaveFolder,
                        left_file_string, right_file_string, confirm_folder)
                    save_index += 1
            self.stcam.close()
        else:
            print("Stereo camera is not defined")

    def close(self):
        """
        Close connection to camera
        """
        if (self.stcam is not None):
            self.stcam.close()


if __name__ == "__main__":
    CAMERA_TYPE_PHOBOS = 0
    CAMERA_TYPE_DEIMOS = 1
    CAMERA_TYPE_PYLON = 2
    CAMERA_TYPE_IMAGE = 3

    camera_type = CAMERA_TYPE_DEIMOS

    stcap = None
    if (camera_type == CAMERA_TYPE_PHOBOS):
        stcap = StereoCapture("Phobos", ["22864917", "22864912"])
    elif (camera_type == CAMERA_TYPE_PYLON):
        stcap = StereoCapture("Pylon", ["22864917", "22864912"])
    elif (camera_type == CAMERA_TYPE_DEIMOS):
        stcap = StereoCapture("Deimos", 0)
    elif (camera_type == CAMERA_TYPE_IMAGE):
        stcap = StereoCapture(
            "Image", ["../SampleData/left.png", "../SampleData/right.png"])
    else:
        print("Invalid camera type.")
        exit()

    stcap.runGui()



#TODO currently linux only so not included in the package untill windows support is added

import numpy
import gi
import time

from collections import namedtuple

gi.require_version("Gst", "1.0")
gi.require_version("Tcam", "0.1")

from gi.repository import Tcam, Gst, GLib, GObject

DeviceInfo = namedtuple("DeviceInfo", "status name identifier connection_type")
CameraProperty = namedtuple("CameraProperty", "status value min max default step type flags category group")

class TIS:
    'The Imaging Source Camera'

    def __init__(self,serial, width, height, framerate, color):
        ''' Constructor
        :param serial: Serial number of the camera to be used.
        :param width: Width of the wanted video format
        :param height: Height of the wanted video format
        :param framerate: Numerator of the frame rate. /1 is added automatically
        :param color: True = 8 bit color, False = 8 bit mono. ToDo: Y16
        :return: none
        '''
        Gst.init([])
        self.height = height
        self.width = width
        self.sample = None
        self.samplelocked = False
        self.newsample = False
        self.img_mat = None
        self.ImageCallback = None

        pixelformat = "BGRx"
        if color is False:
            pixelformat="GRAY8"

        source_name = "source" + serial
        sink_name = "sink" + serial

        p = 'tcambin serial="%s" name=%s ! video/x-raw,format=%s,width=%d,height=%d,framerate=%d/1' % (serial,source_name,pixelformat,width,height,framerate,)
        p += ' ! videoconvert ! appsink name=%s' % (sink_name,)
        # p += ' ! appsink name=sink'

        print(p)
        try:
            self.pipeline = Gst.parse_launch(p)
        except GLib.Error as error:
            print("Error creating pipeline: {0}".format(error))
            raise

        self.pipeline.set_state(Gst.State.READY)
        self.pipeline.get_state(Gst.CLOCK_TIME_NONE)
        # Query a pointer to our source, so we can set properties.
        self.source = self.pipeline.get_by_name(source_name)

        # Query a pointer to the appsink, so we can assign the callback function.
        self.appsink = self.pipeline.get_by_name(sink_name)
        self.appsink.set_property("max-buffers",5)
        self.appsink.set_property("drop",1)
        self.appsink.set_property("emit-signals",1)
        self.appsink.connect('new-sample', self.on_new_buffer)

    def on_new_buffer(self, appsink):
        self.newsample = True
        if self.samplelocked is False:
            try:
                self.sample = appsink.get_property('last-sample')
                if self.ImageCallback is not None:
                    self.__convert_sample_to_numpy()
                    self.ImageCallback(self, *self.ImageCallbackData);

            except GLib.Error as error:
                print("Error on_new_buffer pipeline: {0}".format(error))
                raise
        return False

    def Start_pipeline(self):
        try:
            self.pipeline.set_state(Gst.State.PLAYING)
            self.pipeline.get_state(Gst.CLOCK_TIME_NONE)

        except GLib.Error as error:
            print("Error starting pipeline: {0}".format(error))
            raise

    def __convert_sample_to_numpy(self):
        ''' Convert a GStreamer sample to a numpy array
            Sample code from https://gist.github.com/cbenhagen/76b24573fa63e7492fb6#file-gst-appsink-opencv-py-L34
            The result is in self.img_mat.
        :return:
        '''
        self.samplelocked = True
        buf = self.sample.get_buffer()
        caps = self.sample.get_caps()
        bpp = 4;
        dtype = numpy.uint8
        bla = caps.get_structure(0).get_value('height')
        if( caps.get_structure(0).get_value('format') == "BGRx" ):
            bpp = 4;

        if(caps.get_structure(0).get_value('format') == "GRAY8" ):
            bpp = 1;

        if(caps.get_structure(0).get_value('format') == "GRAY16_LE" ):
            bpp = 1;
            dtype = numpy.uint16

        self.img_mat = numpy.ndarray(
            (caps.get_structure(0).get_value('height'),
             caps.get_structure(0).get_value('width'),
             bpp),
            buffer=buf.extract_dup(0, buf.get_size()),
            dtype=dtype)
        self.newsample = False
        self.samplelocked = False

    def wait_for_image(self,timeout):
        ''' Wait for a new image with timeout
        :param timeout: wait time in second, should be a float number
        :return:
        '''
        tries = 10;
        while tries > 0 and not self.newsample:
            tries -= 1
            time.sleep(float(timeout) / 10.0)

    def Snap_image(self, timeout):
        '''
        Snap an image from stream using a timeout.
        :param timeout: wait time in second, should be a float number. Not used
        :return: bool: True, if we got a new image, otherwise false.
        '''
        if self.ImageCallback is not None:
            print("Snap_image can not be called, if a callback is set.")
            return False

        self.wait_for_image(timeout)
        if( self.sample != None and self.newsample == True):
            self.__convert_sample_to_numpy()
            return True
        return False

    def Get_image(self):
        return self.img_mat

    def Stop_pipeline(self):
        self.pipeline.set_state(Gst.State.PAUSED)
        self.pipeline.set_state(Gst.State.READY)
        self.pipeline.set_state(Gst.State.NULL)

    def List_Properties(self):
        for name in self.source.get_tcam_property_names():
            print( name )

    def Get_Property(self, PropertyName):
        try:
            return CameraProperty(*self.source.get_tcam_property(PropertyName))
        except GLib.Error as error:
            print("Error get Property {0}: {1}",PropertyName, format(error))
            raise

    def Set_Property(self, PropertyName, value):
        try:
            self.source.set_tcam_property(PropertyName,GObject.Value(type(value),value))
        except GLib.Error as error:
            print("Error set Property {0}: {1}",PropertyName, format(error))
            raise

    def Set_Image_Callback(self, function, *data):
        self.ImageCallback = function
        self.ImageCallbackData = data;


import setuptools

with open("../README.md", "r") as fh:
    long_description = fh.read()

with open("../version.txt", "r") as fh:
    version = fh.read()

setuptools.setup(
    name="stereo3d",
    version=version,
    author="Ben Knight",
    author_email="bknight@i3drobotics.com",
    description="Generating 3D data from stereo images.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/i3drobotics/Stereo3D",
    packages=setuptools.find_packages(),
    package_dir={'stereo3d':'stereo3d'},
    package_data={'stereo3d':[
        'stereocapture/coppeliasim/data/MacOS/remoteApi.dylib',
        'stereocapture/coppeliasim/data/Ubuntu16_04/remoteApi.so',
        'stereocapture/coppeliasim/data/Ubuntu18_04/remoteApi.so',
        'stereocapture/coppeliasim/data/Windows/remoteApi.dll',
    ]},
    install_requires=[
        'numpy; python_version == "3.5"',
        'numpy==1.19.3; python_version > "3.5"',
        'opencv-python','pymsgbox','pypylon', 'pyntcloud', 'i3drsgm'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


from stereo3d.stereocapture.cvcapture import CVCapture
import numpy as np
import cv2


class StereoCaptureCVDual():
    def __init__(self, cam):
        """
        Initialisation function for StereoCaptureCVDual class.
        :param cam: camera to use as stereo device. This should be a camera that send the data as a single rgb image where red=left_image and green=right_image.
        :type cam: CVCapture
        """
        self.cam = cam

    def connect(self):
        """
        Connect to camera.
        :returns: success of connection
        :rtype: bool
        """
        res = self.cam.connect()
        return res

    def grab(self):
        """
        Grab images from stereo camera
        :returns: success of capture, image left, image right
        :rtype: bool, numpy.array, numpy.array
        """
        res,image = self.cam.grab()
        image_left = image.copy()
        image_right = image.copy()
        image_left = image_left[:, :, 1]
        image_right = image_right[:, :, 2]
        
        return res,image_left,image_right

    def close(self):
        """
        Close connection to camera
        """
        self.cam.close()


if __name__ == "__main__":
    cam = CVCapture(0)
    stcam = StereoCaptureCVDual(cam)
    stcam.connect()
    while True:
        res, imageL, imageR = stcam.grab()
        if res:
            stereo_image = np.concatenate((imageL, imageR), axis=1)
            stereo_image_resized = cv2.resize(stereo_image, (1280, 480))
            cv2.imshow('Stereo Image', stereo_image_resized)
            k = cv2.waitKey(1)
            if k == ord('q'):
                break
    stcam.close()



import os
import cv2


class CVImageFolderCapture():
    def __init__(self, folder):
        """
        Initialisation function for CVCapture class.
        :param folder: filename of the image to read
        :type folder: string
        """
        self.images = None
        self.image_index = 0
        self.folder = folder

    def connect(self):
        """
        Connect to camera using opencv VideoCaputre
        :returns: success of connection
        :rtype: bool
        """
        self.images = self.load_images_from_folder(self.folder)
        return True

    def load_images_from_folder(self, folder):
        images = []
        for filename in os.listdir(folder):
            img = cv2.imread(
                os.path.join(folder, filename), cv2.IMREAD_GRAYSCALE)
            if img is not None:
                images.append(img)
        return images

    def grab(self):
        """
        Grab image from camera
        :returns: success of capture, image
        :rtype: bool, numpy.array
        """
        image = self.images[self.image_index]
        self.image_index += 1
        if (self.image_index >= len(self.images)):
            self.image_index = 0
        return True, image

    def close(self):
        """
        Close connection
        """


if __name__ == '__main__':
    FOLDER = "pyStereoCapture/SampleData/"
    cam = CVImageFolderCapture(FOLDER)
    cam.connect()
    while True:
        res, image = cam.grab()
        if res:
            image_resized = cv2.resize(image, (640, 480))
            cv2.imshow('Image', image_resized)
            k = cv2.waitKey(1)
            if k == ord('q'):
                break
    cam.close()



import cv2
import numpy as np
from stereo3d.stereocapture.coppeliasim import VREPConnection, VREPVisionSensor
import os


class VREPCapture():
    def __init__(self, api_port, vision_sensor_name):
        """
        Initialisation function for VREPCapture class.
        :param api_port: VREP api port
        :param vision_sensor_name: VREP Vision sensor name
        :type api_port: int
        :type vision_sensor_name: string
        """
        self.camera = None
        self.vrep_connection = None
        self.api_port = api_port
        self.vision_sensor_name = vision_sensor_name

    def connect(self):
        """
        Connect to vision sensor camera using VREP API interface.
        :returns: success
        :rtype: bool
        """
        self.vrep_connection = VREPConnection(self.api_port)
        res, clientID = self.vrep_connection.connect()

        if (res):
            self.camera = VREPVisionSensor(self.vision_sensor_name, clientID)
            self.camera.run_threaded()
        return res

    def grab(self):
        """
        Grab image from camera
        :returns: success of capture, image
        :rtype: bool, numpy.array
        """
        if(self.camera.isInitalised()):
            image = self.camera.cv_image
            image_mono = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            return True, image_mono
        else:
            return False, None

    def close(self):
        """
        Close connection to camera
        """
        self.camera.stop_threaded()


if __name__ == '__main__':
    API_PORT = 20000
    VISION_SENSOR_NAME = "StereoCameraLeft"
    cam = VREPCapture(API_PORT, VISION_SENSOR_NAME)
    cam.connect()
    while True:
        res, image = cam.grab()
        if res:
            image_resized = cv2.resize(image, (640, 480))
            cv2.imshow('Image', image_resized)
            k = cv2.waitKey(1)
            if k == ord('q'):
                break
    cam.close()
